#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from ads.common import logger
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.model.extractor.tensorflow_extractor import TensorflowExtractor
from ads.model.generic_model import FrameworkSpecificModel
from ads.model.model_properties import ModelProperties
from ads.model.serde.model_serializer import TensorflowModelSerializerType
from ads.model.common.utils import DEPRECATE_AS_ONNX_WARNING
from ads.model.serde.common import SERDE


class TensorFlowModel(FrameworkSpecificModel):
    """TensorFlowModel class for estimators from Tensorflow framework.

    Attributes
    ----------
    algorithm: str
        The algorithm of the model.
    artifact_dir: str
        Directory for generate artifact.
    auth: Dict
        Default authentication is set using the `ads.set_auth` API. To override the
        default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create
        an authentication signer to instantiate an IdentityClient object.
    estimator: Callable
        A trained tensorflow estimator/model using Tensorflow.
    framework: str
        "tensorflow", the framework name of the model.
    hyperparameter: dict
        The hyperparameters of the estimator.
    metadata_custom: ModelCustomMetadata
        The model custom metadata.
    metadata_provenance: ModelProvenanceMetadata
        The model provenance metadata.
    metadata_taxonomy: ModelTaxonomyMetadata
        The model taxonomy metadata.
    model_artifact: ModelArtifact
        This is built by calling prepare.
    model_deployment: ModelDeployment
        A ModelDeployment instance.
    model_file_name: str
        Name of the serialized model.
    model_id: str
        The model ID.
    properties: ModelProperties
        ModelProperties object required to save and deploy model.
        For more details, check https://accelerated-data-science.readthedocs.io/en/latest/ads.model.html#module-ads.model.model_properties.
    runtime_info: RuntimeInfo
        A RuntimeInfo instance.
    schema_input: Schema
        Schema describes the structure of the input data.
    schema_output: Schema
        Schema describes the structure of the output data.
    serialize: bool
        Whether to serialize the model to pkl file by default. If False, you need to serialize the model manually,
        save it under artifact_dir and update the score.py manually.
    version: str
        The framework version of the model.

    Methods
    -------
    delete_deployment(...)
        Deletes the current model deployment.
    deploy(..., **kwargs)
        Deploys a model.
    from_model_artifact(uri, model_file_name, artifact_dir, ..., **kwargs)
        Loads model from the specified folder, or zip/tar archive.
    from_model_catalog(model_id, model_file_name, artifact_dir, ..., **kwargs)
        Loads model from model catalog.
    introspect(...)
        Runs model introspection.
    predict(data, ...)
        Returns prediction of input data run against the model deployment endpoint.
    prepare(..., **kwargs)
        Prepare and save the score.py, serialized model and runtime.yaml file.
    reload(...)
        Reloads the model artifact files: `score.py` and the `runtime.yaml`.
    save(..., **kwargs)
        Saves model artifacts to the model catalog.
    summary_status(...)
        Gets a summary table of the current status.
    verify(data, ...)
        Tests if deployment works in local environment.

    Examples
    --------
    >>> from ads.model.framework.tensorflow_model import TensorFlowModel
    >>> import tempfile
    >>> import tensorflow as tf

    >>> mnist = tf.keras.datasets.mnist
    >>> (x_train, y_train), (x_test, y_test) = mnist.load_data()
    >>> x_train, x_test = x_train / 255.0, x_test / 255.0

    >>> tf_estimator = tf.keras.models.Sequential(
    ...                [
    ...                    tf.keras.layers.Flatten(input_shape=(28, 28)),
    ...                    tf.keras.layers.Dense(128, activation="relu"),
    ...                    tf.keras.layers.Dropout(0.2),
    ...                    tf.keras.layers.Dense(10),
    ...                ]
    ...            )
    >>> loss_fn = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)
    >>> tf_estimator.compile(optimizer="adam", loss=loss_fn, metrics=["accuracy"])
    >>> tf_estimator.fit(x_train, y_train, epochs=1)

    >>> tf_model = TensorFlowModel(estimator=tf_estimator,
    ... artifact_dir=tempfile.mkdtemp())
    >>> inference_conda_env = "generalml_p37_cpu_v1"

    >>> tf_model.prepare(inference_conda_env="generalml_p37_cpu_v1", force_overwrite=True)
    >>> tf_model.verify(x_test[:1])
    >>> tf_model.save()
    >>> model_deployment = tf_model.deploy(wait_for_completion=False)
    >>> tf_model.predict(x_test[:1])
    """

    _PREFIX = "tensorflow"
    model_save_serializer_type = TensorflowModelSerializerType

    @runtime_dependency(
        module="tensorflow",
        short_name="tf",
        install_from=OptionalDependency.TENSORFLOW,
    )
    def __init__(
        self,
        estimator: callable,
        artifact_dir: str,
        properties: Optional[ModelProperties] = None,
        auth: Dict = None,
        model_save_serializer: Optional[SERDE] = model_save_serializer_type.TENSORFLOW,
        model_input_serializer: Optional[SERDE] = None,
        **kwargs,
    ):
        """
        Initiates a TensorFlowModel instance.

        Parameters
        ----------
        estimator: callable
            Any model object generated by tensorflow framework
        artifact_dir: str
            Directory for generate artifact.
        properties: (ModelProperties, optional). Defaults to None.
            ModelProperties object required to save and deploy model.
        auth :(Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        model_save_serializer: (SERDE or str, optional). Defaults to None.
            Instance of ads.model.SERDE. Used for serialize/deserialize model.
        model_input_serializer: (SERDE, optional). Defaults to None.
            Instance of ads.model.SERDE. Used for serialize/deserialize data.

        Returns
        -------
        TensorFlowModel
            TensorFlowModel instance.
        """
        super().__init__(
            estimator=estimator,
            artifact_dir=artifact_dir,
            properties=properties,
            auth=auth,
            model_save_serializer=model_save_serializer,
            model_input_serializer=model_input_serializer,
            **kwargs,
        )
        self._extractor = TensorflowExtractor(estimator)
        self.framework = self._extractor.framework
        self.algorithm = self._extractor.algorithm
        self.version = self._extractor.version
        self.hyperparameter = self._extractor.hyperparameter

        self.version = tf.version.VERSION

    def serialize_model(
        self,
        as_onnx: bool = False,
        X_sample: Optional[
            Union[
                Dict,
                str,
                List,
                Tuple,
                np.ndarray,
                pd.core.series.Series,
                pd.core.frame.DataFrame,
            ]
        ] = None,
        force_overwrite: bool = False,
        **kwargs,
    ) -> None:
        """
        Serialize and save Tensorflow model using ONNX or model specific method.

        Parameters
        ----------
        as_onnx: (bool, optional). Defaults to False.
            If set as True, convert into ONNX model.
        X_sample: Union[list, tuple, pd.Series, np.ndarray, pd.DataFrame]. Defaults to None.
            A sample of input data that will be used to generate input schema and detect input_signature.
        force_overwrite: (bool, optional). Defaults to False.
            If set as True, overwrite serialized model if exists.
        **kwargs: optional params used to serialize tensorflow model to onnx,
        including the following:
            input_signature: a tuple or a list of tf.TensorSpec objects). default to None.
            Define the shape/dtype of the input so that model(input_signature) is a valid invocation of the model.
            opset_version: int. Defaults to None. Used for the ONNX model.

        Returns
        -------
        None
            Nothing.
        """

        if as_onnx:
            logger.warning(
                "This approach supports converting tensorflow.keras models to "
                "onnx format. If the defined model includes other tensorflow "
                "modules (e.g., tensorflow.function), please use GenericModel instead."
            )
            logger.warning(DEPRECATE_AS_ONNX_WARNING)
            self.set_model_save_serializer(self.model_save_serializer_type.ONNX)

        super().serialize_model(
            as_onnx=as_onnx,
            force_overwrite=force_overwrite,
            X_sample=X_sample,
            **kwargs,
        )

    @runtime_dependency(
        module="tensorflow",
        short_name="tf",
        install_from=OptionalDependency.TENSORFLOW,
    )
    def _to_tensor(self, data):
        data = tf.convert_to_tensor(data)
        return data



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from __future__ import print_function, absolute_import, division

import pandas as pd

from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.type_discovery import logger
from ads.type_discovery.abstract_detector import AbstractTypeDiscoveryDetector
from ads.type_discovery.typed_feature import DateTimeTypedFeature


class DateTimeDetector(AbstractTypeDiscoveryDetector):
    @runtime_dependency(module="datefinder", install_from=OptionalDependency.DATA)
    def _is_date_time(self, name, values, low_level_type_name):
        if low_level_type_name.startswith("datetime64"):
            return lambda x: x
        else:
            #
            # if the column/feature contains the word "timestamp" then
            #
            if low_level_type_name.startswith("int") and "timestamp" in name.lower():
                # either s (max len 10) on ns (max len) 19
                unit = "s" if values.astype("str").str.len().max() <= 10 else "ns"
                try:
                    pd.to_datetime(values, unit=unit)
                    return lambda x: pd.to_datetime(x, unit=unit)
                except:
                    pass
            if values.dtype == "object":
                try:
                    pd.to_datetime(values, infer_datetime_format=True)
                    datefinder_result = all(
                        [bool(list(datefinder.find_dates(str(x)))) for x in values]
                    )
                    if datefinder_result:
                        return lambda x: pd.to_datetime(x, infer_datetime_format=True)
                except:
                    pass

        return None

    def discover(self, name, series):
        candidates = series.loc[~series.isnull()]
        fn = self._is_date_time(name, candidates.head(500), series.dtype.name)
        if fn:
            logger.debug("column [{}]/[{}] datetime".format(name, series.dtype))
            return DateTimeTypedFeature.build(name, fn(series))

        return False


if __name__ == "__main__":
    dd = DateTimeDetector()
    print(
        dd.discover(
            "date-range",
            pd.Series(pd.date_range(start="1/1/2018", end="1/08/2018", freq="H")),
        )
    )
    print(dd.discover("dates", pd.Series(["12/12/12", "12/12/13", None, "12/12/14"])))
    print(
        dd.discover(
            "dates-with-other-values",
            pd.Series(["12/12/12", "Monday", None, "12/12/14"]),
        )
    )
    print(
        dd.discover(
            "timestamp s",
            pd.Series(
                [
                    978300760,
                    978302109,
                    978301968,
                    978300275,
                    978824291,
                    978302268,
                    978302039,
                ]
            ),
        )
    )
    print(dd.discover("timestamp ns", pd.Series([1490195805433502912])))



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
# This is a placeholder file for container job.



# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from abc import ABC, abstractmethod
from typing import Any, Generator


class Reader(ABC):
    """The Data Reader Interface."""

    @abstractmethod
    def read(self) -> Generator[Any, Any, Any]:
        """The abstract method to read data.

        Yields
        ------
        Generator[Any, Any, Any]
        """
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module with all default warnings provided to user. These are registered to relevant feature
types directly in the feature type files themselves.
"""

import pandas as pd


def missing_values_handler(s: pd.Series) -> pd.DataFrame:
    """
    Warning for > 5 percent missing values (Nans) in series.

    Parameters
    ----------
    s : pd.Series
        Pandas series - column of some feature type.

    Returns
    -------
    pd.Dataframe
        Dataframe with 4 columns 'Warning', 'Message', 'Metric', 'Value'
        and 2 rows, where first row is count of missing values and second is
        percentage of missing values.
    """
    num_missing = s.isna().sum()
    pct_missing = 100 * num_missing / len(s)
    df = pd.DataFrame([], columns=["Warning", "Message", "Metric", "Value"])
    if pct_missing >= 5:
        df.loc[0] = ["missing", f"{num_missing} missing values", "count", num_missing]
        df.loc[1] = [
            "missing",
            f"{pct_missing:.1f}% missing values",
            "percentage",
            round(pct_missing, 2),
        ]
    return df


def skew_handler(s: pd.Series) -> pd.DataFrame:
    """
    Warning if absolute value of skew is greater than 1.

    Parameters
    ----------
    s : pd.Series
        Pandas series - column of some feature type, expects continuous values.

    Returns
    -------
    pd.Dataframe
        Dataframe with 4 columns 'Warning', 'Message', 'Metric', 'Value'
        and 1 rows, which lists skew value of that column.
    """
    series_skew = s.skew()
    df = pd.DataFrame([], columns=["Warning", "Message", "Metric", "Value"])
    if abs(series_skew) > 1:
        df.loc[0] = ["skew", f"{series_skew:.3f} skew", "skew", round(series_skew, 2)]
    return df


def high_cardinality_handler(s: pd.Series) -> pd.DataFrame:
    """
    Warning if number of unique values (including Nan) in series is greater than or equal to 15.

    Parameters
    ----------
    s : pd.Series
        Pandas series - column of some feature type.

    Returns
    -------
    pd.Dataframe
        Dataframe with 4 columns 'Warning', 'Message', 'Metric', 'Value'
        and 1 rows, which lists count of unique values.
    """
    num_unique = s.nunique(dropna=False)
    df = pd.DataFrame([], columns=["Warning", "Message", "Metric", "Value"])
    if num_unique >= 15:
        df.loc[0] = [
            "high-cardinality",
            f"{num_unique} unique values",
            "count",
            num_unique,
        ]
    elif num_unique == len(s):
        df.loc[0] = [
            "high-cardinality",
            f"every value is distinct",
            "count",
            num_unique,
        ]
    return df


def zeros_handler(s: pd.Series) -> pd.DataFrame:
    """
    Warning for greater than 10 percent zeros in series.

    Parameters
    ----------
    s : pd.Series
        Pandas series - column of some feature type.

    Returns
    -------
    pd.Dataframe
        Dataframe with 4 columns 'Warning', 'Message', 'Metric', 'Value'
        and 2 rows, where first row is count of zero values and second is
        percentage of zero values.
    """
    num_zeros = (s == 0).sum()
    pct_missing = 100 * num_zeros / len(s)
    df = pd.DataFrame([], columns=["Warning", "Message", "Metric", "Value"])
    if pct_missing >= 5:
        df.loc[0] = ["zeros", f"{num_zeros} zeros", "count", num_zeros]
        df.loc[1] = [
            "zeros",
            f"{pct_missing:.1f}% zeros",
            "percentage",
            round(pct_missing, 2),
        ]
    return df



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Generator, List

from ads.common import auth as authutil
from ads.common import oci_client
from ads.data_labeling.interface.reader import Reader
from oci.data_labeling_service_dataplane.models import AnnotationSummary, RecordSummary
from oci.exceptions import ServiceError
from oci.pagination import list_call_get_all_results


class ReadRecordsError(Exception):
    def __init__(self, dataset_id: str):
        super().__init__(
            f"Error occurred in attempt to read records of dataset {dataset_id}."
        )


class ReadAnnotationsError(Exception):
    def __init__(self, dataset_id: str):
        super().__init__(
            f"Error occurred in attempt to read annotations of dataset {dataset_id}."
        )


@dataclass
class OCIRecordSummary:
    """The class that representing the labeled record in ADS format.

    Attributes
    ----------
    record: RecordSummary
        OCI RecordSummary.
    annotations: List[AnnotationSummary]
        List of OCI AnnotationSummary.
    """

    record: RecordSummary = None
    annotation: List[AnnotationSummary] = None


class DLSRecordReader(Reader):
    """DLS Record Reader Class that reads records from the cloud into ADS format."""

    def __init__(self, dataset_id: str, compartment_id: str, auth: dict = None):
        """Initiates a DLSRecordReader instance.

        Parameters
        ----------
        dataset_id: str
            The dataset OCID.
        compartment_id: str
            The compartment OCID of the dataset.
        auth: (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        """
        if not dataset_id:
            raise ValueError("The dataset OCID must be specified.")

        if not isinstance(dataset_id, str):
            raise TypeError("The dataset_id must be a string.")

        if not compartment_id:
            raise ValueError("The compartment OCID must be specified.")

        if not isinstance(compartment_id, str):
            raise TypeError("The compartment OCID must be a string.")

        auth = auth or authutil.default_signer()
        self.dataset_id = dataset_id
        self.compartment_id = compartment_id
        self.dls_dp_client = oci_client.OCIClientFactory(**auth).data_labeling_dp

    def _read_records(self):
        try:
            return list_call_get_all_results(
                self.dls_dp_client.list_records,
                self.compartment_id,
                self.dataset_id,
                lifecycle_state="ACTIVE",
            ).data
        except ServiceError:
            raise ReadRecordsError(self.dataset_id)

    def _read_annotations(self):
        try:
            return list_call_get_all_results(
                self.dls_dp_client.list_annotations,
                self.compartment_id,
                self.dataset_id,
                lifecycle_state="ACTIVE",
            ).data
        except ServiceError:
            raise ReadAnnotationsError(self.dataset_id)

    def read(self) -> Generator[OCIRecordSummary, Any, Any]:
        """Reads OCI records.

        Yields
        ------
        OCIRecordSummary
            The OCIRecordSummary instance.
        """
        records = self._read_records()
        annotations = self._read_annotations()

        annotations_map = defaultdict(list)
        for annotation in annotations:
            annotations_map[annotation.record_id].append(annotation)

        for record in records:
            yield OCIRecordSummary(record, annotations_map.get(record.id))



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Dict, List, Optional
from copy import deepcopy

import yaml
from ads.common.serializer import DataClassSerializable

try:
    from yaml import CDumper as dumper
    from yaml import CLoader as loader
except:
    from yaml import Dumper as dumper
    from yaml import Loader as loader

import json
import os
import asteval
from os import path

import fsspec
import yaml
from cerberus import Validator
import sys
from string import Template


SCHEMA_VALIDATOR_NAME = "data_schema.json"
INPUT_OUTPUT_SCHENA_SIZE_LIMIT = 32000
SCHEMA_VERSION = "1.1"
DEFAULT_SCHEMA_VERSION = "1.0"
SCHEMA_KEY = "schema"
SCHEMA_VERSION_KEY = "version"


class SchemaSizeTooLarge(ValueError):
    def __init__(self, size: int):
        super().__init__(
            f"The schema `{size}` bytes and "
            f"the maximum allowable schema size is `{INPUT_OUTPUT_SCHENA_SIZE_LIMIT}` bytes. "
            "Reduce the size of the schema."
        )


class DataSizeTooWide(ValueError):
    def __init__(self, data_col_num: int, max_col_num: int):
        super().__init__(
            f"The data has `{data_col_num}` columns and "
            f"the maximum allowable number of columns is `{max_col_num}`. "
            "Increase allowable number of columns by setting a larger max_col_num. It will take longer time to prepare."
        )


@dataclass(repr=False)
class Expression(DataClassSerializable):
    """
    Expression allows specifying string representation of an expression which can be evaluated by the language corresponding to the value provided in `langauge` attribute

    Default value for language is python

    Parameters
    ----------
    exression: Must use string.Template format for specifying the exression
        type: str
    language: default value is python. It could be any language. `evaluate` method expects the expression to be of type python

    Examples
    --------
    >>> exp = Expression("($x > 10 and $x <100) or ($x < -1 and $x > -500)")
    >>> exp.evaluate(x=500)
    False
    >>> exp.evaluate(x=20)
    True
    >>> exp.evaluate(x=9)
    False
    >>> exp.evaluate(x=-9)
    True
    """

    expression: str
    language: str = "python"

    def evaluate(self, **kwargs):
        if self.language.lower() != "python":
            raise Exception(
                f"Evaluation not supported for language ${self.language}. Supported language: python"
            )

        exp = Template(self.expression)
        final_expression = None
        try:
            final_expression = exp.substitute(kwargs)
        except:
            raise Exception(
                "Error substituting the value into the expression. Check if the variable in the expression matches the parameter names passed to evaluate method"
            )
        aeval = asteval.Interpreter()
        return aeval(final_expression)

    def __post_init__(self):
        if not self.expression or not self.language:
            raise ValueError(
                f"An Expression object requires values for both expression and language"
            )


@dataclass(repr=False)
class Domain(DataClassSerializable):
    """Domain describes the data. It holds following information -
    * stats - Statistics of the data.
    * constraints - List of Expression which defines the constraint for the data.
    * Domain values.

    Examples
    --------

    >>> Domain(values='Rational Numbers', stats={"mean":50, "median":51, "min": 5, "max":100}, constraints=[Expression('$x > 5')])
    constraints:
    - expression: $x > 5
        language: python
    stats:
        max: 100
        mean: 50
        median: 51
        min: 5
    values: Rational Numbers
    """

    values: str = ""
    stats: Dict = field(default_factory=dict)
    constraints: List[Expression] = field(default_factory=list)


@dataclass(repr=False, order=True)
class Attribute(DataClassSerializable):
    """
    Attribute describes the column/feature/element. It holds following information -
    * dtype - Type of data - float, int64, etc. Matches with Pandas dtypes
    * feature_type - Feature type of data - Integer, String, etc. Matches with ads feature types.
    * name - Name of the feature
    * domain - Represented by the Domain class
    * required - Boolean - True or False
    * description - Description about the column/feature
    * order - order of the column/feature in the data

    Examples
    --------

    >>> attr_fruits = Attribute(
    ...     dtype = "category",
    ...     feature_type = "category",
    ...     name = "fruits",
    ...     domain = Domain(values="Apple, Orange, Grapes", stats={"mode": "Orange"}, constraints=[Expression("in ['Apple', 'Orange', 'Grapes']")]),
    ...     required = True,
    ...     description = "Names of fruits",
    ...     order = 0
    ... )
    >>> attr_fruits
    description: Names of fruits
    domain:
        constraints:
        - expression: in ['Apple', 'Orange', 'Grapes']
            language: python
        stats:
            mode: Orange
        values: Apple, Orange, Grapes
    dtype: category
    feature_type: category
    name: fruits
    order: 0
    required: true
    >>> attr_fruits.key
    'fruits'
    """

    sort_index: int = field(init=False, repr=False)
    dtype: str
    feature_type: str
    name: str
    domain: Domain
    required: bool
    description: str
    order: Optional[int] = None

    @property
    def key(self):
        return self.name

    def to_dict(self, **kwargs) -> dict:
        data = super().to_dict(**kwargs)
        data.pop("sort_index", None)
        return data

    def __hash__(self):
        return hash(self.key)

    def __post_init__(self):
        object.__setattr__(self, "sort_index", self.order or 0)


class BaseSchemaLoader(ABC):
    """
    Base Schema Loader which load and validate schema.

    Methods
    -------
    load_schema(self)
        Load and validate schema from a file and return the normalized schema.
    """

    def __init__(self):
        self._schema = None

    def load_schema(self, schema_path):
        """Load and validate schema from a file and return the normalized schema."""
        self._load_schema(schema_path)
        self._normalize()
        return self._validate()

    def _normalize(self):
        self._schema = {key.lower(): value for key, value in self._schema.items()}

    @abstractmethod
    def _load_schema(self, schema_path):
        pass

    def _validate(self):
        """Validate the schema."""
        schema_validator = self._load_schema_validator()
        v = Validator(schema_validator)
        normalized_items = []
        for item in self._schema[SCHEMA_KEY]:
            valid = v.validate(item)
            if not valid:
                new_dict = {"column": item["name"], "error": v.errors}
                raise ValueError(json.dumps(new_dict, indent=2))
            normalized_items.append(v.normalized(item))
        schema_version = self._schema.get(SCHEMA_VERSION_KEY) or DEFAULT_SCHEMA_VERSION
        self._schema = {
            SCHEMA_KEY: normalized_items,
            SCHEMA_VERSION_KEY: schema_version,
        }
        return self._schema

    @staticmethod
    def _load_schema_validator():
        """load the schema validator to validate the schema."""
        with open(
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)), SCHEMA_VALIDATOR_NAME
            )
        ) as schema_file:
            schema_validator = json.load(schema_file)
        return schema_validator


class JsonSchemaLoader(BaseSchemaLoader):
    """
    Json Schema which load and validate schema from json file.

    Methods
    -------
    load_schema(self)
        Load and validate schema from json file and return the normalized schema.

    Examples
    --------
    >>> schema_loader = JsonSchemaLoader()
    >>> schema_dict = schema_loader.load_schema('schema.json')
    >>> schema_dict
    {'Schema': [{'dtype': 'object',
        'feature_type': 'String',
        'name': 'Attrition',
        'domain': {'values': 'String',
            'stats': {'count': 1470, 'unique': 2},
            'constraints': []},
        'required': True,
        'description': 'Attrition'},
        {'dtype': 'int64',
        'feature_type': 'Integer',
        'name': 'Age',
        'domain': {'values': 'Integer',
            'stats': {'count': 1470.0,
            'mean': 37.923809523809524,
            'std': 9.135373489136732,
            'min': 19.0,
            '25%': 31.0,
            '50%': 37.0,
            '75%': 44.0,
            'max': 61.0},
            'constraints': []},
        'required': True,
        'description': 'Age'}]}
    """

    def _load_schema(self, schema_path):
        """Loads and validates schema from a json file."""
        assert os.path.splitext(schema_path)[-1].lower() in [
            ".json"
        ], "Expecting a json format file."
        if not path.exists(schema_path):
            raise FileNotFoundError(f"{schema_path} does not exist")
        with fsspec.open(schema_path, mode="r", encoding="utf8") as f:
            self._schema = json.load(f)


class YamlSchemaLoader(BaseSchemaLoader):
    """
    Yaml Schema which loads and validates schema from a yaml file.

    Methods
    -------
    load_schema(self)
        Loads and validates schema from a yaml file and returns the normalized schema.

    Examples
    --------
    >>> schema_loader = YamlSchemaLoader()
    >>> schema_dict = schema_loader.load_schema('schema.yaml')
    >>> schema_dict
    {'Schema': [{'description': 'Attrition',
        'domain': {'constraints': [],
            'stats': {'count': 1470, 'unique': 2},
            'values': 'String'},
        'dtype': 'object',
        'feature_type': 'String',
        'name': 'Attrition',
        'required': True},
        {'description': 'Age',
        'domain': {'constraints': [],
            'stats': {'25%': 31.0,
            '50%': 37.0,
            '75%': 44.0,
            'count': 1470.0,
            'max': 61.0,
            'mean': 37.923809523809524,
            'min': 19.0,
            'std': 9.135373489136732},
            'values': 'Integer'},
        'dtype': 'int64',
        'feature_type': 'Integer',
        'name': 'Age',
        'required': True}]}
    """

    def _load_schema(self, schema_path):
        """Load and validate schema from yaml file."""
        assert os.path.splitext(schema_path)[-1].lower() in [
            ".yaml",
            ".yml",
        ], "Expecting yaml format file."
        if not path.exists(schema_path):
            raise FileNotFoundError(f"{schema_path} does not exist")
        with open(schema_path, "r") as stream:
            try:
                self._schema = yaml.load(yaml.safe_load(stream), Loader=loader)
            except yaml.YAMLError as exc:
                raise exc


class SchemaFactory:
    """
    Schema Factory.

    Methods
    -------
    register_format(self)
        Register a new type of schema class.
    get_schema(self)
        Get the YamlSchema or JsonSchema based on the format.
    default_schema(cls)
        Construct a SchemaFactory instance and register yaml and json loader.

    Examples
    --------
    >>> factory = SchemaFactory.default_schema()
    >>> schema_loader = factory.get_schema('.json')
    >>> schema_dict = schema_loader.load_schema('schema.json')
    >>> schema = Schema.from_dict(schema_dict)
    >>> schema
    Schema:
    - description: Attrition
    domain:
        constraints: []
        stats:
        count: 1470
        unique: 2
        values: String
    dtype: object
    feature_type: String
    name: Attrition
    required: true
    - description: Age
    domain:
        constraints: []
        stats:
        25%: 31.0
        50%: 37.0
        75%: 44.0
        count: 1470.0
        max: 61.0
        mean: 37.923809523809524
        min: 19.0
        std: 9.135373489136732
        values: Integer
    dtype: int64
    feature_type: Integer
    name: Age
    required: true
    """

    def __init__(self):
        self._creators = {}

    def register_format(self, file_format, creator):
        """Register a new type of schema class."""
        self._creators[file_format] = creator

    def get_schema(self, file_format):
        """Get the YamlSchema or JsonSchema based on the format."""
        creator = self._creators.get(file_format)
        if not creator:
            raise ValueError(
                f"This {file_format} format is not supported. Pass Json or Yaml Files."
            )
        return creator()

    @classmethod
    def default_schema(cls):
        factory = cls()
        factory.register_format(".json", JsonSchemaLoader)
        factory.register_format(".yaml", YamlSchemaLoader)
        factory.register_format(".yml", YamlSchemaLoader)
        return factory


@dataclass(repr=False)
class Schema:
    """
    Schema describes the structure of the data.

    Methods
    -------
    add(self, item: Attribute, replace: bool = False)
        Adds a new attribute item. Replaces existing one if replace flag is True.
    from_dict(self)
        Constructs an instance of Schema from a dictionary.
    from_file(cls, file_path):
        Loads the data schema from a file.
    to_dict(self)
        Serializes the data schema into a dictionary.
    to_yaml(self)
        Serializes the data schema into a YAML.
    to_json(self)
        Serializes the data schema into a json string.
    to_json_file(self)
        Saves the data schema into a json file.
    to_yaml_file(self)
        Save to a yaml file.
    add(self, item: Attribute, replace=False) -> None
        Adds a new attribute item. Replaces existing one if replace flag is True.

    Examples
    --------

    >>> attr_fruits = Attribute(
    ...     dtype = "category",
    ...     feature_type = "category",
    ...     name = "fruits",
    ...     domain = Domain(values="Apple, Orange, Grapes", stats={"mode": "Orange"}, constraints=[Expression("in ['Apple', 'Orange', 'Grapes']")]),
    ...     required = True,
    ...     description = "Names of fruits",
    ...     order = 0,
    ... )
    >>> attr_animals = Attribute(
    ...     dtype = "category",
    ...     feature_type = "category",
    ...     name = "animals",
    ...     domain = Domain(values="Dog, Cat, Python", stats={"mode": "Dog"}, constraints=[Expression("in ['Dog', 'Cat', 'Python']")]),
    ...     required = True,
    ...     description = "Names of animals",
    ...     order = 1,
    ... )
    >>> schema = Schema()
    >>> schema.add(attr_fruits)
    >>> schema.add(attr_animals)
    >>> schema
    schema:
    - description: Names of fruits
    domain:
        constraints:
        - expression: in ['Apple', 'Orange', 'Grapes']
        language: python
        stats:
        mode: Orange
        values: Apple, Orange, Grapes
    dtype: category
    feature_type: category
    name: fruits
    order: 0
    required: true
    - description: Names of animals
    domain:
        constraints:
        - expression: in ['Dog', 'Cat', 'Python']
        language: python
        stats:
        mode: Dog
        values: Dog, Cat, Python
    dtype: category
    feature_type: category
    name: animals
    order: 1
    required: true
    >>> schema.to_dict()
        {'schema': [{'dtype': 'category',
        'feature_type': 'category',
        'name': 'fruits',
        'domain': {'values': 'Apple, Orange, Grapes',
            'stats': {'mode': 'Orange'},
            'constraints': [{'expression': "in ['Apple', 'Orange', 'Grapes']",
            'language': 'python'}]},
        'required': True,
        'description': 'Names of fruits',
        'order': 0},
        {'dtype': 'category',
        'feature_type': 'category',
        'name': 'animals',
        'domain': {'values': 'Dog, Cat, Python',
            'stats': {'mode': 'Dog'},
            'constraints': [{'expression': "in ['Dog', 'Cat', 'Python']",
            'language': 'python'}]},
        'required': True,
        'description': 'Names of animals',
        'order': 1}]}

    """

    _schema: set = field(default_factory=set, init=False)
    _version: str = SCHEMA_VERSION

    def add(self, item: Attribute, replace: bool = False):
        """Adds a new attribute item. Replaces existing one if replace flag is True.

        Overrides the existing one if replace flag is True.

        Parameters
        ----------
        item : Attribute
            The attribute instance of a column/feature/element.
        replace : bool
            Overrides the existing attribute item if replace flag is True.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        ValueError
            If item is already registered and replace flag is False.
        TypeError
            If input data has a wrong format.
        """
        if not isinstance(item, Attribute):
            raise TypeError("Argument must be an instance of the class Attribute.")
        if not replace and item.key in self.keys:
            raise ValueError(
                f"The key {item.key} already exists. Use `replace=True` to overwrite."
            )
        self._schema.discard(item)
        self._schema.add(item)

    @property
    def keys(self) -> list:
        """Returns all registered Attribute keys.

        Returns
        -------
        Tuple[str]
            The list of Attribute keys.
        """
        return tuple(item.key for item in self)

    @classmethod
    def from_dict(cls, schema: dict):
        """Constructs an instance of Schema from a dictionary.

        Parameters
        ----------
        schema : dict
            Data schema in dictionary format.

        Returns
        -------
        Schema
            An instance of Schema.
        """
        sc = cls()

        if schema == {}:
            return sc
        if not isinstance(schema, dict):
            raise TypeError("schema has to be of dictionary type.")

        schema = {key.lower(): value for key, value in deepcopy(schema).items()}
        for item in schema[SCHEMA_KEY]:
            domain = Domain(**item["domain"])
            domain.constraints = []
            for constraint in item["domain"]["constraints"]:
                domain.constraints.append(Expression(**constraint))
            item["domain"] = domain
            sc.add(Attribute(**item))
        return sc

    @classmethod
    def from_json(cls, schema: str):
        """Constructs an instance of Schema from a Json.

        Parameters
        ----------
        schema : str
            Data schema in Json format.

        Returns
        -------
        Schema
            An instance of Schema.
        """
        return Schema.from_dict(json.loads(schema))

    @classmethod
    def from_file(cls, file_path: str):
        """Loads the data schema from a file.

        Parameters
        ----------
        file_path : str
            File Path to load the data schema.

        Returns
        -------
        Schema
            An instance of Schema.
        """
        file_format = os.path.splitext(file_path)[-1]
        schema_loader = SchemaFactory.default_schema().get_schema(file_format)
        return cls.from_dict(schema_loader.load_schema(file_path))

    def to_dict(self):
        """Serializes data schema into a dictionary.

        Returns
        -------
        dict
            The dictionary representation of data schema.
        """
        return {
            SCHEMA_KEY: [item.to_dict() for item in self],
            SCHEMA_VERSION_KEY: self._version,
        }

    def to_yaml(self):
        """Serializes the data schema into a YAML.
        Returns
        -------
        str
            The yaml representation of data schema.
        """
        return yaml.dump(self.to_dict(), Dumper=dumper)

    def to_json(self):
        """Serializes the data schema into a json string.
        Returns
        -------
        str
            The json representation of data schema.
        """
        return json.dumps(self.to_dict()).replace("NaN", "null")

    def to_json_file(self, file_path):
        """Saves the data schema into a json file.

        Parameters
        ----------
        file_path : str
            File Path to store the schema in json format.

        Returns
        -------
        None
            Nothing.
        """
        directory = os.path.expanduser(os.path.dirname(file_path))
        basename = os.path.expanduser(os.path.basename(file_path))
        assert os.path.splitext(file_path)[-1].lower() in [
            ".json"
        ], f"The file `{basename}` is not a valid JSON file. The `{file_path}` must have the extension .json."
        if directory and not os.path.exists(directory):
            try:
                os.mkdir(directory)
            except:
                raise Exception(f"Error creating the directory.")
        with open(os.path.join(directory, basename), "w") as json_file:
            json.dump(self.to_dict(), json_file)

    def to_yaml_file(self, file_path):
        """Saves the data schema into a yaml file.
        Parameters
        ----------
        file_path : str
            File Path to store the schema in yaml format.

        Returns
        -------
        None
            Nothing.
        """
        assert os.path.splitext(file_path)[-1] in [
            ".yaml",
            ".yml",
        ], "The `file_path` must have the extension .yaml or .yml."
        directory = os.path.expanduser(os.path.dirname(file_path))
        basename = os.path.expanduser(os.path.basename(file_path))
        if directory and not os.path.exists(directory):
            try:
                os.mkdir(directory)
            except:
                raise Exception(f"Error creating the directory.")
        with open(os.path.join(directory, basename), "w") as yaml_file:
            yaml.dump(self.to_yaml(), yaml_file, default_flow_style=True)

    def validate_size(self) -> bool:
        """Validates schema size.

        Validates the size of schema. Throws an error if the size of the schema
        exceeds expected value.

        Returns
        -------
        bool
            True if schema does not exceeds the size limit.

        Raises
        ------
        SchemaSizeTooLarge
            If the size of the schema exceeds expected value.
        """
        if sys.getsizeof(self.to_yaml()) > INPUT_OUTPUT_SCHENA_SIZE_LIMIT:
            raise SchemaSizeTooLarge(sys.getsizeof(self.to_yaml()))
        return True

    def validate_schema(self):
        """Validate the schema."""
        schema_validator = BaseSchemaLoader._load_schema_validator()
        v = Validator(schema_validator)
        for item in self.to_dict()[SCHEMA_KEY]:
            valid = v.validate(item)
            if not valid:
                new_dict = {"column": item["name"], "error": v.errors}
                raise ValueError(json.dumps(new_dict, indent=2))
        return True

    def __getitem__(self, key: str):
        if key is None or key == "":
            raise ValueError(f"The key `{key}` must not be empty.")
        if not isinstance(key, str):
            TypeError(f"The key `{key}` must be a string.")
        for item in self._schema:
            if item.key == key:
                return item
        raise ValueError(f"The key {key} is not found.")

    def __repr__(self):
        return self.to_yaml()

    def __iter__(self):
        return sorted(self._schema).__iter__()

    def __len__(self):
        return len(self._schema)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
from __future__ import annotations

import datetime
import logging
import os
import time
import traceback
import uuid
from io import DEFAULT_BUFFER_SIZE
from string import Template
from typing import Any, Dict, List, Optional, Union

import fsspec
import oci.data_science
import oci.util as oci_util
import yaml
from oci.data_science.models import JobInfrastructureConfigurationDetails
from oci.exceptions import ServiceError
from ads.common import utils
from ads.common.oci_datascience import DSCNotebookSession, OCIDataScienceMixin
from ads.common.oci_logging import OCILog
from ads.common.oci_resource import ResourceNotFoundError
from ads.jobs.builders.infrastructure.base import Infrastructure, RunInstance
from ads.jobs.builders.infrastructure.dsc_job_runtime import (
    DataScienceJobRuntimeManager,
)
from ads.jobs.builders.infrastructure.utils import get_value
from ads.jobs.builders.runtimes.artifact import Artifact
from ads.jobs.builders.runtimes.container_runtime import ContainerRuntime
from ads.jobs.builders.runtimes.python_runtime import GitPythonRuntime

logger = logging.getLogger(__name__)

SLEEP_INTERVAL = 3
WAIT_SECONDS_AFTER_FINISHED = 90


class DSCJob(OCIDataScienceMixin, oci.data_science.models.Job):
    """Represents an OCI Data Science Job
    This class contains all attributes of the oci.data_science.models.Job.
    The main purpose of this class is to link the oci.data_science.models.Job model and the related client methods.
    Mainly, linking the Job model (payload) to Create/Update/Get/List/Delete methods.

    A DSCJob can be initialized by unpacking a the properties stored in a dictionary (payload):

    .. code-block:: python

        job_properties = {
            "display_name": "my_job",
            "job_infrastructure_configuration_details": {"shape_name": "VM.MY_SHAPE"}
        }
        job = DSCJob(**job_properties)

    The properties can also be OCI REST API payload, in which the keys are in camel format.

    .. code-block:: python

        job_payload = {
            "projectId": "<project_ocid>",
            "compartmentId": "<compartment_ocid>",
            "displayName": "<job_name>",
            "jobConfigurationDetails": {
                "jobType": "DEFAULT",
                "commandLineArguments": "pos_arg1 pos_arg2 --key1 val1 --key2 val2",
                "environmentVariables": {
                    "KEY1": "VALUE1",
                    "KEY2": "VALUE2",
                    # User specifies conda env via env var
                    "CONDA_ENV_TYPE" : "service",
                    "CONDA_ENV_SLUG" : "mlcpuv1"
                }
            },
            "jobInfrastructureConfigurationDetails": {
                "jobInfrastructureType": "STANDALONE",
                "shapeName": "VM.Standard.E3.Flex",
                "jobShapeConfigDetails": {
                    "memoryInGBs": 16,
                    "ocpus": 1
                },
                "blockStorageSizeInGBs": "100",
                "subnetId": "<subnet_ocid>"
            }
        }
        job = DSCJob(**job_payload)
    """

    DEFAULT_INFRA_TYPE = (
        JobInfrastructureConfigurationDetails.JOB_INFRASTRUCTURE_TYPE_ME_STANDALONE
    )

    CONST_DEFAULT_BLOCK_STORAGE_SIZE = 50

    def __init__(self, artifact: Union[str, Artifact] = None, **kwargs) -> None:
        """Initialize a DSCJob object.

        Parameters
        ----------
        artifact: str or Artifact
            Job artifact, which can be a path or an Artifact object. Defaults to None.
        kwargs:
            Same as kwargs in oci.data_science.models.Job.
            Keyword arguments are passed into OCI Job model to initialize the properties.

        """
        self._artifact = artifact

        super().__init__(**kwargs)
        if not self.job_configuration_details:
            self.job_configuration_details = {
                "jobType": "DEFAULT",
            }
        if not self.job_infrastructure_configuration_details:
            self.job_infrastructure_configuration_details = {}

    @property
    def artifact(self) -> Union[str, Artifact]:
        """Job artifact.

        Returns
        -------
        str or Artifact
            When creating a job, this be a path or an Artifact object.
            When loading the job from OCI, this will be the filename of the job artifact.
        """
        if self.id and self._artifact is None:
            try:
                res = self.client.head_job_artifact(self.id)
                content = res.headers.get("content-disposition")
                if content and "filename=" in content:
                    self._artifact = content.split("filename=", 1)[-1]
            except ServiceError:
                self._artifact = ""
        return self._artifact

    @artifact.setter
    def artifact(self, artifact: Union[str, Artifact]):
        """Sets the job artifact."""
        self._artifact = artifact

    def _load_infra_from_notebook(self, nb_config):
        """Loads the infrastructure configuration from notebook configuration."""
        infra = self.job_infrastructure_configuration_details
        nb_shape_config_details = oci_util.to_dict(
            getattr(nb_config, "notebook_session_shape_config_details", None) or {}
        )
        if isinstance(infra, dict):
            shape_name = infra.get("shapeName", nb_config.shape)

            # Ignore notebook shape config details if shape names do not match.
            if shape_name != nb_config.shape:
                nb_shape_config_details = {}

            infra_type = infra.get("jobInfrastructureType")
            block_storage = infra.get(
                "blockStorageSizeInGBs", nb_config.block_storage_size_in_gbs
            )
            subnet_id = infra.get(
                "subnetId",
                nb_config.subnet_id
                if infra_type
                != JobInfrastructureConfigurationDetails.JOB_INFRASTRUCTURE_TYPE_ME_STANDALONE
                else None,
            )
            job_shape_config_details = infra.get("jobShapeConfigDetails", {})
            memory_in_gbs = job_shape_config_details.get(
                "memoryInGBs", nb_shape_config_details.get("memory_in_gbs")
            )
            ocpus = job_shape_config_details.get(
                "ocpus", nb_shape_config_details.get("ocpus")
            )
        else:
            shape_name = (
                infra.shape_name
                if getattr(infra, "shape_name", None)
                else nb_config.shape
            )
            # Ignore notebook shape config details if shape names do not match.
            if shape_name != nb_config.shape:
                nb_shape_config_details = {}

            infra_type = getattr(infra, "job_infrastructure_type", None)

            block_storage = (
                infra.block_storage_size_in_gbs
                if getattr(infra, "block_storage_size_in_gbs", None)
                else nb_config.block_storage_size_in_gbs
            )
            subnet_id = (
                infra.subnet_id
                if getattr(infra, "subnet_id", None)
                else (
                    nb_config.subnet_id
                    if infra_type
                    != JobInfrastructureConfigurationDetails.JOB_INFRASTRUCTURE_TYPE_ME_STANDALONE
                    else None
                )
            )
            job_shape_config_details = oci_util.to_dict(
                getattr(infra, "job_shape_config_details", {}) or {}
            )
            memory_in_gbs = job_shape_config_details.get(
                "memory_in_gbs", nb_shape_config_details.get("memory_in_gbs")
            )
            ocpus = job_shape_config_details.get(
                "ocpus", nb_shape_config_details.get("ocpus")
            )

        self.job_infrastructure_configuration_details = {
            "jobInfrastructureType": infra_type,
            "shapeName": shape_name,
            "blockStorageSizeInGBs": block_storage,
        }
        # ADS does not provide explicit API for setting infrastructure type.
        # If subnet is configured, the type will be set to STANDALONE,
        # otherwise ME_STANDALONE
        if subnet_id:
            self.job_infrastructure_configuration_details.update(
                {
                    "subnetId": subnet_id,
                    "jobInfrastructureType": JobInfrastructureConfigurationDetails.JOB_INFRASTRUCTURE_TYPE_STANDALONE,
                }
            )
        else:
            self.job_infrastructure_configuration_details.update(
                {
                    "jobInfrastructureType": self.DEFAULT_INFRA_TYPE,
                }
            )

        # Specify shape config details
        if memory_in_gbs or ocpus:
            self.job_infrastructure_configuration_details.update(
                {
                    "jobShapeConfigDetails": {
                        "memoryInGBs": memory_in_gbs,
                        "ocpus": ocpus,
                    }
                }
            )

    def load_properties_from_env(self) -> None:
        """Loads default properties from the environment"""
        if "NB_SESSION_OCID" in os.environ:
            try:
                nb_session = DSCNotebookSession.from_ocid(os.environ["NB_SESSION_OCID"])
            except Exception:
                logger.debug("Failed to load config from notebook.")
                logger.debug(traceback.format_exc())
                # If there is an error loading the notebook infra configurations.
                # Ignore it by setting nb_session to None
                # This will skip loading the default configure.
                nb_session = None
            if nb_session:
                nb_config = getattr(
                    nb_session,
                    "notebook_session_config_details",
                    getattr(nb_session, "notebook_session_configuration_details", None),
                )

                if nb_config:
                    self._load_infra_from_notebook(nb_config)
                if self.project_id is None:
                    self.project_id = nb_session.project_id
        super().load_properties_from_env()

    def load_defaults(self) -> DSCJob:
        self.load_properties_from_env()
        if not self.job_infrastructure_configuration_details:
            self.job_infrastructure_configuration_details = {}
        # Convert the dict to JobInfrastructureConfigurationDetails object
        if isinstance(self.job_infrastructure_configuration_details, dict):
            # Default networking
            if not self.job_infrastructure_configuration_details.get(
                "jobInfrastructureType"
            ):
                self.job_infrastructure_configuration_details[
                    "jobInfrastructureType"
                ] = self.DEFAULT_INFRA_TYPE
            self.job_infrastructure_configuration_details = self.deserialize(
                self.job_infrastructure_configuration_details,
                JobInfrastructureConfigurationDetails.__name__,
            )

        # Default block storage size
        if not self.job_infrastructure_configuration_details.block_storage_size_in_gbs:
            self.job_infrastructure_configuration_details.block_storage_size_in_gbs = (
                self.CONST_DEFAULT_BLOCK_STORAGE_SIZE
            )
        return self

    def _create_with_oci_api(self) -> None:
        res = self.client.create_job(
            self.to_oci_model(oci.data_science.models.CreateJobDetails)
        )
        self.update_from_oci_model(res.data)
        if self.lifecycle_state == "ACTIVE":
            return
        try:
            if issubclass(self.artifact.__class__, Artifact):
                with self.artifact as artifact:
                    self.upload_artifact(artifact.path)
            else:
                self.upload_artifact()
        except Exception as ex:
            # Delete the job if upload artifact is failed.
            self.delete()
            raise ex

    def create(self) -> DSCJob:
        """Create the job on OCI Data Science platform

        Returns
        -------
        DSCJob
            The DSCJob instance (self), which allows chaining additional method.

        """
        if not self.display_name:
            if self.artifact:
                timestamp = datetime.datetime.utcnow().strftime("%Y-%m-%d-%H:%M.%S")
                self.display_name = (
                    os.path.basename(str(self.artifact)).split(".")[0] + f"-{timestamp}"
                )
            else:
                # Set default display_name if not specified - randomly generated easy to remember name generated
                self.display_name = utils.get_random_name_for_resource()
        try:
            self.load_defaults()
        except Exception:
            logger.exception("Failed to load default properties.")
        # Check compartment ID and project ID before calling the OCI API
        if not self.compartment_id:
            raise ValueError("Specify compartment ID for data science job.")
        if not self.project_id:
            raise ValueError("Specify project ID for data science job.")
        self._create_with_oci_api()
        return self

    def update(self) -> DSCJob:
        """Updates the Data Science Job."""
        raise NotImplementedError("Updating Job is not supported at the moment.")

    def delete(self) -> DSCJob:
        """Deletes the job and the corresponding job runs.

        Returns
        -------
        DSCJob
            The DSCJob instance (self), which allows chaining additional method.

        """
        runs = self.run_list()
        for run in runs:
            run.delete()
        self.client.delete_job(self.id)
        return self

    def upload_artifact(self, artifact_path: str = None) -> DSCJob:
        """Uploads the job artifact to OCI

        Parameters
        ----------
        artifact_path : str, optional
            Local path to the job artifact file to be uploaded, by default None.
            If artifact_path is None, the path in self.artifact will be used.

        Returns
        -------
        DSCJob
            The DSCJob instance (self), which allows chaining additional method.

        """
        if not artifact_path:
            artifact_path = self.artifact
        with fsspec.open(artifact_path, "rb") as f:
            self.client.create_job_artifact(
                self.id,
                f,
                content_disposition=f"attachment; filename={os.path.basename(artifact_path)}",
            )
        return self

    def download_artifact(self, artifact_path: str) -> DSCJob:
        """Downloads the artifact from OCI

        Parameters
        ----------
        artifact_path : str
            Local path to store the job artifact.

        Returns
        -------
        DSCJob
            The DSCJob instance (self), which allows chaining additional method.

        """
        res = self.client.get_job_artifact_content(self.id)
        with open(artifact_path, "wb") as f:
            for chunk in res.data.iter_content(chunk_size=DEFAULT_BUFFER_SIZE * 16):
                f.write(chunk)
        return self

    def run_list(self, **kwargs) -> list[DataScienceJobRun]:
        """Lists the runs of this job.

        Parameters
        ----------
        **kwargs :
            Keyword arguments to te passed into the OCI list_job_runs() for filtering the job runs.

        Returns
        -------
        list
            A list of DSCJobRun objects

        """
        items = oci.pagination.list_call_get_all_results(
            self.client.list_job_runs, self.compartment_id, job_id=self.id, **kwargs
        ).data
        return [DataScienceJobRun.from_oci_model(item) for item in items]

    def run(self, **kwargs) -> DataScienceJobRun:
        """Runs the job

        Parameters
        ----------
        **kwargs :
            Keyword arguments for initializing a Data Science Job Run.
            The keys can be any keys in supported by OCI JobConfigurationDetails and JobRun, including:
            * hyperparameter_values: dict(str, str)
            * environment_variables: dict(str, str)
            * command_line_arguments: str
            * maximum_runtime_in_minutes: int
            * display_name: str

        If display_name is not specified, it will be generated as "<JOB_NAME>-run-<TIMESTAMP>".

        Returns
        -------
        DSCJobRun
            An instance of DSCJobRun, which can be used to monitor the job run.

        """
        if not self.id:
            self.create()

        swagger_types = (
            oci.data_science.models.DefaultJobConfigurationDetails().swagger_types.keys()
        )
        config_kwargs = {}
        keys = list(kwargs.keys())
        for key in keys:
            if key in swagger_types:
                config_kwargs[key] = kwargs.pop(key)

        # remove timestamp from the job name (added in default names, when display_name not specified by user)
        if self.display_name:
            try:
                datetime.datetime.strptime(self.display_name[-19:], "%Y-%m-%d-%H:%M.%S")
                self.display_name = self.display_name[:-20]
            except ValueError:
                pass

        job_attrs = dict(
            project_id=self.project_id,
            display_name=self.display_name
            + "-run-"
            + datetime.datetime.now().strftime("%Y-%m-%d-%H:%M.%S"),
            job_id=self.id,
            compartment_id=self.compartment_id,
        )

        for key, val in job_attrs.items():
            if not kwargs.get(key):
                kwargs[key] = val

        if config_kwargs:
            config_kwargs["jobType"] = "DEFAULT"
            config_override = kwargs.get("job_configuration_override_details", {})
            config_override.update(config_kwargs)
            kwargs["job_configuration_override_details"] = config_override

        wait = kwargs.pop("wait", False)
        run = DataScienceJobRun(**kwargs).create()
        if wait:
            return run.watch()
        return run

    @classmethod
    def from_ocid(cls, ocid) -> DSCJob:
        """Gets a job by OCID

        Parameters
        ----------
        ocid : str
            The OCID of the job.

        Returns
        -------
        DSCJob
            An instance of DSCJob.

        """
        return super().from_ocid(ocid)


class DataScienceJobRun(
    OCIDataScienceMixin, oci.data_science.models.JobRun, RunInstance
):
    """Represents a Data Science Job run"""

    _DETAILS_LINK = (
        "https://console.{region}.oraclecloud.com/data-science/job-runs/{id}"
    )

    TERMINAL_STATES = [
        oci.data_science.models.JobRun.LIFECYCLE_STATE_SUCCEEDED,
        oci.data_science.models.JobRun.LIFECYCLE_STATE_FAILED,
        oci.data_science.models.JobRun.LIFECYCLE_STATE_CANCELED,
        oci.data_science.models.JobRun.LIFECYCLE_STATE_DELETED,
    ]

    def create(self) -> DataScienceJobRun:
        """Creates a job run"""
        self.load_properties_from_env()
        res = self.client.create_job_run(
            self.to_oci_model(oci.data_science.models.CreateJobRunDetails)
        )
        self.update_from_oci_model(res.data)
        return self

    @property
    def status(self) -> str:
        """Lifecycle status

        Returns
        -------
        str
            Status in a string.
        """
        return self.lifecycle_state

    @property
    def log_id(self) -> str:
        """The log ID from OCI logging service containing the logs from the job run."""
        if not self.log_details:
            return None
        return self.log_details.log_id

    @property
    def log_group_id(self) -> str:
        """The log group ID from OCI logging service containing the logs from the job run."""
        if not self.log_details:
            return None
        return self.log_details.log_group_id

    @property
    def logging(self) -> OCILog:
        """The OCILog object containing the logs from the job run"""
        if not self.log_id:
            raise ValueError("Log OCID is not specified for this job run.")
        # Specifying log group ID when initializing OCILog can reduce the number of API calls.
        return OCILog(id=self.log_id, log_group_id=self.log_details.log_group_id)

    @staticmethod
    def _format_log(message: str, date_time: datetime.datetime) -> dict:
        """Formats a message as log record with datetime.
        This is used to add additional logs to show job run status change.

        Parameters
        ----------
        message : str
            Log message.
        date_time : datetime or str
            Timestamp for the message

        Returns
        -------
        dict
            log record as a dictionary, including id, time and message as keys.
        """
        if isinstance(date_time, datetime.datetime):
            date_time = date_time.strftime("%Y-%m-%dT%H:%M:%S.000Z")
        return {
            "id": str(uuid.uuid4()),
            "message": message,
            "time": date_time,
        }

    def logs(self, limit: int = None) -> list:
        """Gets the logs of the job run.

        Parameters
        ----------
        limit : int, optional
            Limit the number of logs to be returned.
            Defaults to None. All logs will be returned.

        Returns
        -------
        list
            A list of log records. Each log record is a dictionary with the following keys: id, time, message.
        """
        if self.time_accepted:
            log_messages = self.logging.tail(
                source=self.id, limit=limit, time_start=self.time_accepted
            )
        else:
            log_messages = []
        if self.time_started:
            log_messages.insert(
                0, self._format_log("Job Run STARTED", self.time_started)
            )
        if self.time_accepted:
            log_messages.insert(
                0, self._format_log("Job Run ACCEPTED", self.time_accepted)
            )
        if self.time_finished:
            log_messages.append(
                self._format_log("Job Run FINISHED", self.time_finished)
            )
        return log_messages

    def _job_run_status_text(self) -> str:
        details = f", {self.lifecycle_details}" if self.lifecycle_details else ""
        return f"Job Run {self.lifecycle_state}" + details

    def _check_and_print_status(self, prev_status) -> str:
        status = self._job_run_status_text()
        if status != prev_status:
            if self.lifecycle_state in self.TERMINAL_STATES and self.time_finished:
                timestamp = self.time_finished.strftime("%Y-%m-%d %H:%M:%S")
            else:
                timestamp = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
            print(f"{timestamp} - {status}")
        return status

    def watch(
        self,
        interval: float = SLEEP_INTERVAL,
        wait: float = WAIT_SECONDS_AFTER_FINISHED,
    ) -> DataScienceJobRun:
        """Watches the job run until it finishes.
        Before the job start running, this method will output the job run status.
        Once the job start running,
        the logs will be streamed until the job is success, failed or cancelled.

        Parameters
        ----------
        interval : float
            Time interval in seconds between each request to update the logs.
            Defaults to 3 (seconds).
        wait : float
            Time in seconds to keep updating the logs after the job run finished.
            It may take some time for logs to appear in OCI logging service
            after the job run is finished.
            Defaults to 90 (seconds).

        """

        def stop_condition():
            """Stops the log streaming once the job is in a terminal state."""
            self.sync()
            if self.lifecycle_state not in self.TERMINAL_STATES:
                return False
            # Stop if time_finished is not available.
            if not self.time_finished:
                return True
            # Stop only if time_finished is over 2 minute ago.
            # This is for the time delay between job run stopped and the logs appear in oci logging.
            if (
                datetime.datetime.now(self.time_finished.tzinfo)
                - datetime.timedelta(seconds=wait)
                > self.time_finished
            ):
                return True
            return False

        if not self.log_id and not self.log_group_id:
            print(
                "Logging is not configured for the job. Watch() will only show job status."
            )

        print(f"Job OCID: {self.job.id}")
        print(f"Job Run OCID: {self.id}")

        status = ""
        while not stop_condition():
            status = self._check_and_print_status(status)
            # Break and stream logs if job has log ID and started.
            # Otherwise, keep watching the status until job terminates.
            if self.time_started and self.log_id:
                break
            time.sleep(interval)

        if self.log_id and self.time_accepted:
            count = self.logging.stream(
                source=self.id,
                interval=interval,
                stop_condition=stop_condition,
                time_start=self.time_accepted,
            )
            if not count:
                print(
                    "No logs in the last 14 days. Please set time_start to see older logs."
                )

        self._check_and_print_status(status)

        return self

    def cancel(self) -> DataScienceJobRun:
        """Cancels a job run
        This method will wait for the job run to be canceled before returning.

        Returns
        -------
        self
            The job run instance.
        """
        self.client.cancel_job_run(self.id)
        while self.lifecycle_state != "CANCELED":
            self.sync()
            time.sleep(SLEEP_INTERVAL)
        return self

    def __repr__(self) -> str:
        """Displays the object as YAML."""
        return self.to_yaml()

    def to_yaml(self) -> str:
        """Serializes the object into YAML string.

        Returns
        -------
        str
            YAML stored in a string.
        """
        # Here the job YAML is used as the base for the job run
        job_dict = self.job.to_dict()

        # Update infrastructure from job run
        run_dict = self.to_dict()
        infra_specs = [
            run_dict,
            run_dict.get("jobInfrastructureConfigurationDetails", {}),
            run_dict.get("logDetails", {}),
        ]
        for infra_spec in infra_specs:
            for key in infra_spec:
                if key in job_dict["spec"]["infrastructure"]["spec"]:
                    job_dict["spec"]["infrastructure"]["spec"][key] = infra_spec[key]

        # Update runtime from job run
        from ads.jobs import Job

        job = Job.from_dict(job_dict)
        envs = job.runtime.envs
        run_config_override = run_dict.get("jobConfigurationOverrideDetails", {})
        envs.update(run_config_override.get("environmentVariables", {}))
        job.runtime.with_environment_variable(**envs)
        if run_config_override.get("commandLineArguments"):
            job.runtime.set_spec(
                "args",
                run_config_override.get("commandLineArguments"),
            )

        # Update kind, id and name
        run_dict = job.to_dict()
        run_dict["kind"] = "jobRun"
        run_dict["spec"]["id"] = self.id
        run_dict["spec"]["name"] = self.display_name
        return yaml.safe_dump(run_dict)

    @property
    def job(self):
        """The job instance of this run.

        Returns
        -------
        Job
            An ADS Job instance
        """
        from ads.jobs import Job

        return Job.from_datascience_job(self.job_id)

    def download(self, to_dir):
        """Downloads files from job run output URI to local.

        Parameters
        ----------
        to_dir : str
            Local directory to which the files will be downloaded to.

        Returns
        -------
        DataScienceJobRun
            The job run instance (self)
        """
        self.job.download(to_dir)
        return self


# This is for backward compatibility
DSCJobRun = DataScienceJobRun


class DataScienceJob(Infrastructure):
    """Represents the OCI Data Science Job infrastructure.

    To configure the infrastructure for a Data Science Job::

        infrastructure = (
            DataScienceJob()
            # Configure logging for getting the job run outputs.
            .with_log_group_id("<log_group_ocid>")
            # Log resource will be auto-generated if log ID is not specified.
            .with_log_id("<log_ocid>")
            # If you are in an OCI data science notebook session,
            # the following configurations are not required.
            # Configurations from the notebook session will be used as defaults.
            .with_compartment_id("<compartment_ocid>")
            .with_project_id("<project_ocid>")
            .with_subnet_id("<subnet_ocid>")
            .with_shape_name("VM.Standard.E3.Flex")
            # Shape config details are applicable only for the flexible shapes.
            .with_shape_config_details(memory_in_gbs=16, ocpus=1)
            # Minimum/Default block storage size is 50 (GB).
            .with_block_storage_size(50)
        )

    """

    CONST_PROJECT_ID = "projectId"
    CONST_COMPARTMENT_ID = "compartmentId"
    CONST_DISPLAY_NAME = "displayName"
    CONST_JOB_TYPE = "jobType"
    CONST_JOB_INFRA = "jobInfrastructureType"
    CONST_SHAPE_NAME = "shapeName"
    CONST_BLOCK_STORAGE = "blockStorageSize"
    CONST_SUBNET_ID = "subnetId"
    CONST_SHAPE_CONFIG_DETAILS = "shapeConfigDetails"
    CONST_MEMORY_IN_GBS = "memoryInGBs"
    CONST_OCPUS = "ocpus"
    CONST_LOG_ID = "logId"
    CONST_LOG_GROUP_ID = "logGroupId"

    attribute_map = {
        CONST_PROJECT_ID: "project_id",
        CONST_COMPARTMENT_ID: "compartment_id",
        CONST_DISPLAY_NAME: "display_name",
        CONST_JOB_TYPE: "job_type",
        CONST_JOB_INFRA: "job_infrastructure_type",
        CONST_SHAPE_NAME: "shape_name",
        CONST_BLOCK_STORAGE: "block_storage_size",
        CONST_SUBNET_ID: "subnet_id",
        CONST_SHAPE_CONFIG_DETAILS: "shape_config_details",
        CONST_LOG_ID: "log_id",
        CONST_LOG_GROUP_ID: "log_group_id",
    }

    shape_config_details_attribute_map = {
        CONST_MEMORY_IN_GBS: "memory_in_gbs",
        CONST_OCPUS: "ocpus",
    }

    payload_attribute_map = {
        CONST_PROJECT_ID: "project_id",
        CONST_COMPARTMENT_ID: "compartment_id",
        CONST_DISPLAY_NAME: "display_name",
        CONST_JOB_TYPE: "job_configuration_details.job_type",
        CONST_JOB_INFRA: "job_infrastructure_configuration_details.job_infrastructure_type",
        CONST_SHAPE_NAME: "job_infrastructure_configuration_details.shape_name",
        CONST_BLOCK_STORAGE: "job_infrastructure_configuration_details.block_storage_size_in_gbs",
        CONST_SUBNET_ID: "job_infrastructure_configuration_details.subnet_id",
        CONST_SHAPE_CONFIG_DETAILS: "job_infrastructure_configuration_details.job_shape_config_details",
        CONST_LOG_ID: "job_log_configuration_details.log_id",
        CONST_LOG_GROUP_ID: "job_log_configuration_details.log_group_id",
    }

    snake_to_camel_map = {
        v.split(".", maxsplit=1)[-1]: k for k, v in payload_attribute_map.items()
    }

    @staticmethod
    def standardize_spec(spec):
        if not spec:
            return {}

        attribute_map = {
            **DataScienceJob.attribute_map,
            **DataScienceJob.shape_config_details_attribute_map,
        }
        snake_to_camel_map = {v: k for k, v in attribute_map.items()}

        for key in list(spec.keys()):
            if key not in attribute_map and key in snake_to_camel_map:
                value = spec.pop(key)
                if isinstance(value, dict):
                    spec[snake_to_camel_map[key]] = DataScienceJob.standardize_spec(
                        value
                    )
                else:
                    spec[snake_to_camel_map[key]] = value
        return spec

    def __init__(self, spec: Dict = None, **kwargs) -> None:
        """Initializes a data science job infrastructure

        Parameters
        ----------
        spec : dict, optional
            Object specification, by default None
        kwargs: dict
            Specification as keyword arguments.
            If spec contains the same key as the one in kwargs, the value from kwargs will be used.
        """
        self.standardize_spec(spec)
        self.standardize_spec(kwargs)
        super().__init__(spec=spec, **kwargs)
        if not self.job_type:
            self.with_job_type("DEFAULT")
        self.dsc_job = DSCJob()
        self.runtime = None
        self._name = None

    @property
    def name(self) -> str:
        """Display name of the job"""
        if self.dsc_job:
            self._name = self.dsc_job.display_name
        return self._name

    @name.setter
    def name(self, value: str):
        """Sets the display name of the job

        Parameters
        ----------
        value : str
            The display name of the job
        """
        self._name = value
        if self.dsc_job:
            self.dsc_job.display_name = value

    @property
    def job_id(self) -> Optional[str]:
        """The OCID of the job"""
        if self.dsc_job:
            return self.dsc_job.id
        return None

    @property
    def status(self) -> Optional[str]:
        """Status of the job.

        Returns
        -------
        str
            Status of the job.
        """
        if self.dsc_job:
            return self.dsc_job.lifecycle_state
        return None

    def with_project_id(self, project_id: str) -> DataScienceJob:
        """Sets the project OCID

        Parameters
        ----------
        project_id : str
            The project OCID

        Returns
        -------
        DataScienceJob
            The DataScienceJob instance (self)

        """
        return self.set_spec(self.CONST_PROJECT_ID, project_id)

    @property
    def project_id(self) -> Optional[str]:
        """Project OCID"""
        return self.get_spec(self.CONST_PROJECT_ID)

    def with_compartment_id(self, compartment_id: str) -> DataScienceJob:
        """Sets the compartment OCID

        Parameters
        ----------
        compartment_id : str
            The compartment OCID

        Returns
        -------
        DataScienceJob
            The DataScienceJob instance (self)

        """
        return self.set_spec(self.CONST_COMPARTMENT_ID, compartment_id)

    @property
    def compartment_id(self) -> Optional[str]:
        """The compartment OCID"""
        return self.get_spec(self.CONST_COMPARTMENT_ID)

    def with_job_type(self, job_type: str) -> DataScienceJob:
        """Sets the job type

        Parameters
        ----------
        job_type : str
            Job type as string

        Returns
        -------
        DataScienceJob
            The DataScienceJob instance (self)

        """
        return self.set_spec(self.CONST_JOB_TYPE, job_type)

    @property
    def job_type(self) -> Optional[str]:
        """Job type"""
        return self.get_spec(self.CONST_JOB_TYPE)

    def with_job_infrastructure_type(self, infrastructure_type: str) -> DataScienceJob:
        """Sets the job infrastructure type

        Parameters
        ----------
        infrastructure_type : str
            Job infrastructure type as string

        Returns
        -------
        DataScienceJob
            The DataScienceJob instance (self)

        """
        return self.set_spec(self.CONST_JOB_INFRA, infrastructure_type)

    @property
    def job_infrastructure_type(self) -> Optional[str]:
        """Job infrastructure type"""
        return self.get_spec(self.CONST_JOB_INFRA)

    def with_shape_name(self, shape_name: str) -> DataScienceJob:
        """Sets the shape name for running the job

        Parameters
        ----------
        shape_name : str
            Shape name

        Returns
        -------
        DataScienceJob
            The DataScienceJob instance (self)

        """
        return self.set_spec(self.CONST_SHAPE_NAME, shape_name)

    @property
    def shape_name(self) -> Optional[str]:
        """Shape name"""
        return self.get_spec(self.CONST_SHAPE_NAME)

    def with_block_storage_size(self, size_in_gb: int) -> DataScienceJob:
        """Sets the block storage size in GB

        Parameters
        ----------
        size_in_gb : int
            Block storage size in GB

        Returns
        -------
        DataScienceJob
            The DataScienceJob instance (self)

        """
        return self.set_spec(self.CONST_BLOCK_STORAGE, size_in_gb)

    @property
    def block_storage_size(self) -> int:
        """Block storage size for the job"""
        return self.get_spec(self.CONST_BLOCK_STORAGE)

    def with_subnet_id(self, subnet_id: str) -> DataScienceJob:
        """Sets the subnet ID

        Parameters
        ----------
        subnet_id : str
            Subnet ID

        Returns
        -------
        DataScienceJob
            The DataScienceJob instance (self)

        """
        return self.set_spec(self.CONST_SUBNET_ID, subnet_id)

    @property
    def subnet_id(self) -> str:
        """Subnet ID"""
        return self.get_spec(self.CONST_SUBNET_ID)

    def with_shape_config_details(
        self, memory_in_gbs: float, ocpus: float, **kwargs: Dict[str, Any]
    ) -> DataScienceJob:
        """Sets the details for the job run shape configuration.
        Specify only when a flex shape is selected.
        For example `VM.Standard.E3.Flex` allows the memory_in_gbs and cpu count to be specified.

        Parameters
        ----------
        memory_in_gbs: float
            The size of the memory in GBs.
        ocpus: float
            The OCPUs count.
        kwargs
            Additional keyword arguments.

        Returns
        -------
        DataScienceJob
            The DataScienceJob instance (self)
        """
        return self.set_spec(
            self.CONST_SHAPE_CONFIG_DETAILS,
            {
                self.CONST_OCPUS: ocpus,
                self.CONST_MEMORY_IN_GBS: memory_in_gbs,
                **kwargs,
            },
        )

    @property
    def shape_config_details(self) -> Dict:
        """The details for the job run shape configuration."""
        return self.get_spec(self.CONST_SHAPE_CONFIG_DETAILS)

    def with_log_id(self, log_id: str) -> DataScienceJob:
        """Sets the log OCID for the data science job.
        If log ID is specified, setting the log group ID (with_log_group_id()) is not strictly needed.
        ADS will look up the log group ID automatically.
        However, this may require additional permission,
        and the look up may not be available for newly created log group.
        Specifying both log ID (with_log_id()) and log group ID (with_log_group_id())
        can avoid such lookup and speed up the job creation.

        Parameters
        ----------
        log_id : str
            Log resource OCID.

        Returns
        -------
        DataScienceJob
            The DataScienceJob instance (self)
        """
        return self.set_spec(self.CONST_LOG_ID, log_id)

    @property
    def log_id(self) -> str:
        """Log OCID for the data science job.

        Returns
        -------
        str
            Log OCID
        """
        return self.get_spec(self.CONST_LOG_ID)

    def with_log_group_id(self, log_group_id: str) -> DataScienceJob:
        """Sets the log group OCID for the data science job.
        If log group ID is specified but log ID is not,
        a new log resource will be created automatically for each job run to store the logs.

        Parameters
        ----------
        log_group_id : str
            Log Group OCID

        Returns
        -------
        DataScienceJob
            The DataScienceJob instance (self)
        """
        return self.set_spec(self.CONST_LOG_GROUP_ID, log_group_id)

    @property
    def log_group_id(self) -> str:
        """Log group OCID of the data science job

        Returns
        -------
        str
            Log group OCID
        """
        return self.get_spec(self.CONST_LOG_GROUP_ID)

    def _prepare_log_config(self) -> dict:
        if not self.log_group_id and not self.log_id:
            return None
        # Look up log group ID if only the log ID is specified
        if self.log_id and not self.log_group_id:
            try:
                log_obj = OCILog.from_ocid(self.log_id)
            except ResourceNotFoundError as exc:
                raise ResourceNotFoundError(
                    f"Unable to determine log group ID for Log ({self.log_id})."
                    " The log resource may not exist or You may not have the required permission."
                    " Try to avoid this by specifying the log group ID."
                ) from exc
            self.with_log_group_id(log_obj.log_group_id)

        if self.log_group_id and not self.log_id:
            enable_auto_log_creation = True
        else:
            enable_auto_log_creation = False

        log_config = {
            "enable_logging": True,
            "enable_auto_log_creation": enable_auto_log_creation,
        }
        if self.log_id:
            log_config["log_id"] = self.log_id

        if self.log_group_id:
            log_config["log_group_id"] = self.log_group_id
        return log_config

    def _update_from_dsc_model(
        self, dsc_job: oci.data_science.models.Job, overwrite: bool = True
    ) -> DataScienceJob:
        """Update the properties from an OCI data science job model.

        Parameters
        ----------
        dsc_job: oci.data_science.models.Job
            An OCI data science job model.

        overwrite: bool
            Whether to overwrite the existing values.
            If this is set to False, only the empty/None properties will be updated.

        Returns
        -------
        DataScienceJob
            The DataScienceJob instance (self)
        """
        sub_level = {
            self.CONST_SHAPE_CONFIG_DETAILS: self.shape_config_details_attribute_map
        }
        self.dsc_job = dsc_job

        for infra_attr, dsc_attr in self.payload_attribute_map.items():
            value = get_value(dsc_job, dsc_attr)
            if not value:
                continue
            if infra_attr not in sub_level:
                if overwrite or not self._spec.get(infra_attr):
                    self._spec[infra_attr] = value
            else:
                sub_spec = self._spec.get(infra_attr, {})
                self._spec[infra_attr] = {}
                for sub_infra_attr, sub_dsc_attr in sub_level[infra_attr].items():
                    sub_value = get_value(value, sub_dsc_attr)
                    if not sub_value:
                        continue
                    if overwrite or not sub_spec.get(sub_infra_attr):
                        sub_spec[sub_infra_attr] = sub_value
                if sub_spec:
                    self._spec[infra_attr] = sub_spec
        return self

    def _update_job_infra(self, dsc_job: DSCJob) -> DataScienceJob:
        """Updates the job infrastructure from a DSCJob object.

        Parameters
        ----------
        dsc_job : DSCJob
            A DSCJob instance.

        Returns
        -------
        DataScienceJob
            The DataScienceJob instance (self)

        """
        attr_map = {
            self.CONST_JOB_INFRA: "jobInfrastructureType",
            self.CONST_SHAPE_NAME: "shapeName",
            self.CONST_SUBNET_ID: "subnetId",
            self.CONST_BLOCK_STORAGE: "blockStorageSizeInGBs",
            self.CONST_SHAPE_CONFIG_DETAILS: "jobShapeConfigDetails",
        }

        if not dsc_job.job_infrastructure_configuration_details:
            dsc_job.job_infrastructure_configuration_details = {}

        for snake_attr, camel_attr in attr_map.items():
            value = self.get_spec(snake_attr)
            if value:
                dsc_job.job_infrastructure_configuration_details[camel_attr] = value

        if dsc_job.job_infrastructure_configuration_details.get("subnetId"):
            dsc_job.job_infrastructure_configuration_details[
                "jobInfrastructureType"
            ] = JobInfrastructureConfigurationDetails.JOB_INFRASTRUCTURE_TYPE_STANDALONE
        return self

    def build(self) -> DataScienceJob:
        self.dsc_job.load_defaults()
        self._update_from_dsc_model(self.dsc_job, overwrite=False)
        return self

    def create(self, runtime, **kwargs) -> DataScienceJob:
        """Creates a job with runtime.

        Parameters
        ----------
        runtime : Runtime
            An ADS job runtime.

        Returns
        -------
        DataScienceJob
            The DataScienceJob instance (self)

        """
        if not runtime:
            raise ValueError("Set a valid runtime.")
        payload = DataScienceJobRuntimeManager(self).translate(runtime)
        # Add infra properties to payload
        for attr in ["project_id", "compartment_id"]:
            if getattr(self, attr):
                payload[attr] = getattr(self, attr)

        if self.name:
            display_name = Template(self.name).safe_substitute(runtime.envs)
        elif isinstance(runtime, GitPythonRuntime) or isinstance(
            runtime, ContainerRuntime
        ):
            display_name = utils.get_random_name_for_resource()
        else:
            display_name = None

        payload["display_name"] = display_name
        payload["job_log_configuration_details"] = self._prepare_log_config()

        self.dsc_job = DSCJob(**payload)
        # Set Job infra to user values after DSCJob initialized the defaults
        self._update_job_infra(self.dsc_job)
        self.dsc_job.create()
        # Update the model from infra after job creation.
        self._update_from_dsc_model(self.dsc_job)
        return self

    def run(
        self, name=None, args=None, env_var=None, freeform_tags=None, wait=False
    ) -> DataScienceJobRun:
        """Runs a job on OCI Data Science job

        Parameters
        ----------
        name : str, optional
            The name of the job run, by default None.
        args : str, optional
            Command line arguments for the job run, by default None.
        env_var : dict, optional
            Environment variable for the job run, by default None
        freeform_tags : dict, optional
            Freeform tags for the job run, by default None
        wait : bool, optional
            Indicate if this method should wait for the run to finish before it returns, by default False.

        Returns
        -------
        DataScienceJobRun
            A Data Science Job Run instance.

        """
        # Runtime in the infrastructure will be None if the job is not created.
        if not self.runtime:
            raise RuntimeError(
                "Job is not created. Call create() to create the job first."
            )
        tags = self.runtime.freeform_tags
        if not tags:
            tags = {}
        if freeform_tags:
            tags.update(freeform_tags)

        if name:
            envs = self.runtime.envs
            if env_var:
                envs.update(env_var)
            name = Template(name).safe_substitute(envs)

        return self.dsc_job.run(
            display_name=name,
            command_line_arguments=args,
            environment_variables=env_var,
            freeform_tags=tags,
            wait=wait,
        )

    def delete(self) -> None:
        """Deletes a job"""
        self.dsc_job.delete()

    def run_list(self, **kwargs) -> List[DataScienceJobRun]:
        """Gets a list of job runs.

        Parameters
        ----------
        **kwargs :
            Keyword arguments for filtering the job runs.
            These arguments will be passed to OCI API.


        Returns
        -------
        List[DSCJobRun]:
            A list of job runs.

        """
        return self.dsc_job.run_list(**kwargs)

    @classmethod
    def from_dsc_job(cls, dsc_job: DSCJob) -> DataScienceJob:
        """Initialize a DataScienceJob instance from a DSCJob

        Parameters
        ----------
        dsc_job : DSCJob
            An instance of DSCJob

        Returns
        -------
        DataScienceJob
            An instance of DataScienceJob

        """
        instance = cls()
        instance._update_from_dsc_model(dsc_job)
        instance.runtime = DataScienceJobRuntimeManager(instance).extract(dsc_job)
        return instance

    @classmethod
    def from_id(cls, job_id: str) -> DataScienceJob:
        """Gets an existing job using Job OCID

        Parameters
        ----------
        job_id : str
            Job OCID


        Returns
        -------
        DataScienceJob
            An instance of DataScienceJob

        """
        return cls.from_dsc_job(DSCJob.from_ocid(job_id))

    @classmethod
    def list_jobs(cls, compartment_id: str = None, **kwargs) -> List[DataScienceJob]:
        """Lists all jobs in a compartment.

        Parameters
        ----------
        compartment_id : str, optional
            The compartment ID for running the jobs, by default None.
            This is optional in a OCI Data Science notebook session.
            If this is not specified, the compartment ID of the notebook session will be used.
        **kwargs :
            Keyword arguments to be passed into OCI list_jobs API for filtering the jobs.

        Returns
        -------
        List[DataScienceJob]
            A list of DataScienceJob object.

        """
        return [
            cls.from_dsc_job(job)
            for job in DSCJob.list_resource(compartment_id, **kwargs)
        ]

    @classmethod
    def instance_shapes(cls, compartment_id: str = None, **kwargs) -> list:
        """Lists the supported shapes for running jobs in a compartment.

        Parameters
        ----------
        compartment_id : str, optional
            The compartment ID for running the jobs, by default None.
            This is optional in a OCI Data Science notebook session.
            If this is not specified, the compartment ID of the notebook session will be used.

        Returns
        -------
        list
            A list of oci.data_science.models.JobShapeSummary objects
            containing the information of the supported shapes.

        Examples
        --------
        To get a list of shape names::

            shapes = DataScienceJob.fast_launch_shapes(
                compartment_id=os.environ["PROJECT_COMPARTMENT_OCID"]
            )
            shape_names = [shape.name for shape in shapes]

        """
        shapes = oci.pagination.list_call_get_all_results(
            DSCJob.init_client().list_job_shapes,
            DSCJob.check_compartment_id(compartment_id),
            **kwargs,
        ).data
        return shapes

    @classmethod
    def fast_launch_shapes(cls, compartment_id: str = None, **kwargs) -> list:
        """Lists the supported fast launch shapes for running jobs in a compartment.

        Parameters
        ----------
        compartment_id : str, optional
            The compartment ID for running the jobs, by default None.
            This is optional in a OCI Data Science notebook session.
            If this is not specified, the compartment ID of the notebook session will be used.

        Returns
        -------
        list
            A list of oci.data_science.models.FastLaunchJobConfigSummary objects
            containing the information of the supported shapes.

        Examples
        --------
        To get a list of shape names::

            shapes = DataScienceJob.fast_launch_shapes(
                compartment_id=os.environ["PROJECT_COMPARTMENT_OCID"]
            )
            shape_names = [shape.shape_name for shape in shapes]

        """
        shapes = oci.pagination.list_call_get_all_results(
            DSCJob.init_client().list_fast_launch_job_configs,
            DSCJob.check_compartment_id(compartment_id),
            **kwargs,
        ).data
        return shapes



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


class DatasetType:
    """DatasetType class which contains all the dataset
    types that data labeling service supports.
    """

    TEXT = "TEXT"
    IMAGE = "IMAGE"
    DOCUMENT = "DOCUMENT"


class AnnotationType:
    """AnnotationType class which contains all the annotation
    types that data labeling service supports.
    """

    SINGLE_LABEL = "SINGLE_LABEL"
    MULTI_LABEL = "MULTI_LABEL"
    ENTITY_EXTRACTION = "ENTITY_EXTRACTION"
    BOUNDING_BOX = "BOUNDING_BOX"


class Formats:
    """Common formats class which contains all the common
    formats that are supported to convert to.
    """

    SPACY = "spacy"
    YOLO = "yolo"


FORMATS_TO_ANNOTATION_TYPE = {
    Formats.SPACY: AnnotationType.ENTITY_EXTRACTION,
    Formats.YOLO: AnnotationType.BOUNDING_BOX,
}


ANNOTATION_TYPE_TO_FORMATS = {
    AnnotationType.ENTITY_EXTRACTION: Formats.SPACY,
    AnnotationType.BOUNDING_BOX: Formats.YOLO,
}



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import re
from typing import Callable


class UDF:
    @staticmethod
    def from_regex(regex: str) -> Callable:
        def function(content):
            match = re.match(regex, content)
            if match:
                if len(match.groups()) == 0:
                    return [match.group(0)]
                else:
                    return match.groups()

        return function



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
"""
APIs to interact with Oracle's Model Deployment service.

There are three main classes: ModelDeployment, ModelDeploymentDetails, ModelDeployer.

One creates a ModelDeployment and deploys it under the umbrella of the ModelDeployer class. This way
multiple ModelDeployments can be unified with one ModelDeployer. The ModelDeployer class also serves
as the interface to all the deployments. ModelDeploymentDetails holds information about the particular
details of a particular deployment, such as how many instances, etc. In this way multiple, independent
ModelDeployments with the same details can be created using the ModelDeployer class.

Examples
--------
>>> from model_deploy.model_deployer import ModelDeployer, ModelDeploymentDetails
>>> deployer = ModelDeployer("model_dep_conf.yaml")
>>> deployment_properties = ModelDeploymentProperties(
...             'ocid1.datasciencemodel.ocn.reg.xxxxxxxxxxxxxxxxxxxxxxxxx')
...                 .with_prop('display_name', "My model display name")
...                 .with_prop("project_id", project_id)
...                 .with_prop("compartment_id", compartment_id)
...                 .with_instance_configuration(
...                     config={"INSTANCE_SHAPE":"VM.Standard.E3.Flex",
...                             "INSTANCE_COUNT":"1",
...                             "bandwidth_mbps":10,
...                             "memory_in_gbs":10,
...                             "ocpus":2}
...                  ).build()
>>> deployment_info = deployer.deploy(deployment_properties,
...             max_wait_time=600, poll_interval=15)
>>> print(deployment_info.model_deployment_id)
>>> print(deployment_info.workflow_req_id)
>>> print(deployment_info.url)
>>> deployer.list_deployments() # Optionally pass in a status
"""
from typing import Dict, Union
import warnings

import oci.pagination
import pandas as pd
from ads.common.auth import default_signer
from oci.data_science.data_science_client import DataScienceClient
from oci.data_science import DataScienceClientCompositeOperations

from .common import utils
from .common.utils import OCIClientManager, State
from .model_deployment import DEFAULT_POLL_INTERVAL, DEFAULT_WAIT_TIME, ModelDeployment
from .model_deployment_properties import ModelDeploymentProperties


class ModelDeployer:
    """ModelDeployer is the class responsible for deploying the ModelDeployment

    Attributes
    ----------
    config : dict
        ADS auth dictionary for OCI authentication.
    ds_client : DataScienceClient
        data science client
    ds_composite_client : DataScienceCompositeClient
        composite data science client

    Methods
    -------
    deploy(model_deployment_details, **kwargs)
        Deploy the model specified by `model_deployment_details`.
    get_model_deployment(model_deployment_id:str)
        Get the ModelDeployment specified by `model_deployment_id`.
    get_model_deployment_state(model_deployment_id)
        Get the state of the current deployment specified by id.
    delete(model_deployment_id, **kwargs)
        Remove the model deployment specified by the id or Model Deployment Object
    list_deployments(status)
        lists the model deployments associated with current compartment and data
        science client
    show_deployments(status)
        shows the deployments filtered by `status` in a Dataframe
    """

    def __init__(self, config: dict = None, ds_client: DataScienceClient = None):
        """Initializes model deployer.

        Parameters
        ----------
        config : dict, optional
            ADS auth dictionary for OCI authentication.
            This can be generated by calling ads.common.auth.api_keys() or ads.common.auth.resource_principal().
            If this is None, ads.common.default_signer(client_kwargs) will be used.

         ds_client: oci.data_science.data_science_client.DataScienceClient
            The Oracle DataScience client.
        """
        if config:
            warnings.warn(
                "`config` will be deprecated in 3.0.0 and will be ignored now. Please use `ads.set_auth()` to config the auth information."
            )

        self.ds_client = None
        self.ds_composite_client = None

        if ds_client:
            self.ds_client = ds_client
            self.ds_composite_client = DataScienceClientCompositeOperations(
                self.ds_client
            )

        if not (self.ds_client != None and self.ds_composite_client != None):
            self.client_manager = OCIClientManager()
            self.ds_client = self.client_manager.ds_client
            self.ds_composite_client = self.client_manager.ds_composite_client

    def deploy(
        self,
        properties: Union[ModelDeploymentProperties, Dict] = None,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
        **kwargs,
    ) -> ModelDeployment:
        """Deploys a model.

        Parameters
        ----------
        properties : ModelDeploymentProperties or dict
            Properties to deploy the model.
            Properties can be None when kwargs are used for specifying properties.
        wait_for_completion : bool
            Flag set for whether to wait for deployment to complete before proceeding.
            Optional, defaults to True.
        max_wait_time : int
            Maximum amount of time to wait in seconds. Optional, defaults to 1200.
            Negative value implies infinite wait time.
        poll_interval : int
            Poll interval in seconds. Optional, defaults to 30.
        kwargs :
            Keyword arguments for initializing ModelDeploymentProperties.
            See ModelDeploymentProperties() for details.

        Returns
        -------
        ModelDeployment
            A ModelDeployment instance.
        """
        warnings.warn(
            "Make sure to provide `compartment_id`, `project_id`, `instance_count`, `instance_shape`, `memory_in_gbs` and `ocpus` either through `kwargs` or `properties`"
        )
        model_deployment = ModelDeployment(
            properties,
            **kwargs,
        )
        return model_deployment.deploy(
            wait_for_completion, max_wait_time, poll_interval
        )

    def deploy_from_model_uri(
        self,
        model_uri: str,
        properties: Union[ModelDeploymentProperties, Dict] = None,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
        **kwargs,
    ) -> ModelDeployment:
        """Deploys a model.

        Parameters
        ----------
        model_uri : str
            uri to model files, can be local or in cloud storage
        properties : ModelDeploymentProperties or dict
            Properties to deploy the model.
            Properties can be None when kwargs are used for specifying properties.
        wait_for_completion : bool
            Flag set for whether to wait for deployment to complete before proceeding.
            Defaults to True
        max_wait_time : int
            Maximum amount of time to wait in seconds (Defaults to 1200).
            Negative implies infinite wait time.
        poll_interval : int
            Poll interval in seconds (Defaults to 30).
        kwargs :
            Keyword arguments for initializing ModelDeploymentProperties

        Returns
        -------
        ModelDeployment
            A ModelDeployment instance

        """
        warnings.warn(
            "Make sure to provide `compartment_id`, `project_id`, `instance_count`, `instance_shape`, `memory_in_gbs` and `ocpus` either through `kwargs` or `properties`"
        )
        if properties:
            model_id = self.client_manager.prepare_artifact(
                model_uri=model_uri, properties=properties
            )
            properties.model_deployment_configuration_details.model_configuration_details.model_id = (
                model_id
            )
        else:
            model_id = self.client_manager.prepare_artifact(
                model_uri=model_uri, properties=kwargs
            )
            kwargs["model_id"] = model_id

        return self.deploy(
            properties,
            wait_for_completion=wait_for_completion,
            max_wait_time=max_wait_time,
            poll_interval=poll_interval,
            **kwargs,
        )

    def update(
        self,
        model_deployment_id: str,
        properties: ModelDeploymentProperties = None,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
        **kwargs,
    ) -> ModelDeployment:
        """Updates an existing model deployment.

        Parameters
        ----------
        model_deployment_id : str
            Model deployment OCID.
        properties : ModelDeploymentProperties
            An instance of ModelDeploymentProperties or dict to initialize the ModelDeploymentProperties.
            Defaults to None.
        wait_for_completion : bool
            Flag set for whether to wait for deployment to complete before proceeding.
            Defaults to True.
        max_wait_time : int
            Maximum amount of time to wait in seconds (Defaults to 1200).
        poll_interval : int
            Poll interval in seconds (Defaults to 30).
        kwargs :
            Keyword arguments for initializing ModelDeploymentProperties.


        Returns
        -------
        ModelDeployment
            A ModelDeployment instance

        """

        model_deployment = self.get_model_deployment(model_deployment_id)
        # Deployment properties will be refreshed within model_deployment.update() when update is done.
        return model_deployment.update(
            properties,
            wait_for_completion,
            max_wait_time=max_wait_time,
            poll_interval=poll_interval,
            **kwargs,
        )

    def get_model_deployment(self, model_deployment_id: str) -> ModelDeployment:
        """Gets a ModelDeployment by OCID.

        Parameters
        ----------
        model_deployment_id : str
            Model deployment OCID

        Returns
        -------
        ModelDeployment
            A ModelDeployment instance

        """

        try:
            oci_model_deployment_object = self.ds_client.get_model_deployment(
                model_deployment_id
            ).data
            model_deployment_object = ModelDeployment(
                oci_model_deployment_object,
            )
            model_deployment_object.set_spec(
                model_deployment_object.CONST_ID, oci_model_deployment_object.id
            )
            return model_deployment_object.sync()
        except Exception as e:
            utils.get_logger().error(
                "Getting model deployment failed with error: %s", e
            )
            raise e

    def get_model_deployment_state(self, model_deployment_id: str) -> State:
        """Gets the state of a deployment specified by OCID

        Parameters
        ----------
        model_deployment_id : str
            Model deployment OCID

        Returns
        -------
        str
            The state of the deployment
        """

        model_deployment = self.get_model_deployment(model_deployment_id)
        return model_deployment.state

    def delete(
        self,
        model_deployment_id,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
    ) -> ModelDeployment:
        """Deletes the model deployment specified by OCID.

        Parameters
        ----------
        model_deployment_id : str
            Model deployment OCID.
        wait_for_completion : bool
            Wait for deletion to complete. Defaults to True.
        max_wait_time : int
            Maximum amount of time to wait in seconds (Defaults to 600).
            Negative implies infinite wait time.
        poll_interval : int
            Poll interval in seconds (Defaults to 60).

        Returns
        -------
        A ModelDeployment instance that was deleted

        """
        try:
            model_deployment_object = self.get_model_deployment(model_deployment_id)
            return model_deployment_object.delete(
                wait_for_completion, max_wait_time, poll_interval
            )
        except Exception as e:
            utils.get_logger().error(
                "Deleting model deployment failed with error: %s", format(e)
            )
            raise e

    def list_deployments(self, status=None, compartment_id=None, **kwargs) -> list:
        """Lists the model deployments associated with current compartment and data science client

        Parameters
        ----------
        status : str
            Status of deployment. Defaults to None.
        compartment_id : str
            Target compartment to list deployments from.
            Defaults to the compartment set in the environment variable "NB_SESSION_COMPARTMENT_OCID".
            If "NB_SESSION_COMPARTMENT_OCID" is not set, the root compartment ID will be used.
            An ValueError will be raised if root compartment ID cannot be determined.
        kwargs :
            The values are passed to oci.data_science.DataScienceClient.list_model_deployments.

        Returns
        -------
        list
            A list of ModelDeployment objects.

        Raises
        ------
        ValueError
            If compartment_id is not specified and cannot be determined from the environment.

        """
        if not compartment_id:
            compartment_id = self.client_manager.default_compartment_id()
        if not compartment_id:
            raise ValueError(
                "Unable to determine compartment ID from environment. Specify compartment_id."
            )

        if isinstance(status, State):
            status = status.name

        if status is not None:
            kwargs["lifecycle_state"] = status

        # https://oracle-cloud-infrastructure-python-sdk.readthedocs.io/en/latest/api/pagination.html#module-oci.pagination
        deployments = oci.pagination.list_call_get_all_results(
            self.ds_client.list_model_deployments, compartment_id, **kwargs
        ).data

        result = []
        for deployment in deployments:
            model_deployment = ModelDeployment(deployment)
            model_deployment.set_spec(model_deployment.CONST_ID, deployment.id)
            result.append(model_deployment.sync())

        return result

    def show_deployments(
        self,
        status=None,
        compartment_id=None,
    ) -> pd.DataFrame:
        """Returns the model deployments associated with current compartment and data science client
            as a Dataframe that can be easily visualized

        Parameters
        ----------
        status : str
            Status of deployment. Defaults to None.
        compartment_id : str
            Target compartment to list deployments from.
            Defaults to the compartment set in the environment variable "NB_SESSION_COMPARTMENT_OCID".
            If "NB_SESSION_COMPARTMENT_OCID" is not set, the root compartment ID will be used.
            An ValueError will be raised if root compartment ID cannot be determined.

        Returns
        -------
        DataFrame
            pandas Dataframe containing information about the ModelDeployments

        Raises
        ------
        ValueError
            If compartment_id is not specified and cannot be determined from the environment.
        """
        if not compartment_id:
            compartment_id = self.client_manager.default_compartment_id()
        if not compartment_id:
            raise ValueError(
                "Unable to determine compartment ID from environment. Specify compartment_id."
            )

        if type(status) == str or status == None:
            status = State._from_str(status)
        model_deployments = self.ds_client.list_model_deployments(compartment_id).data
        display = pd.DataFrame()
        ids, urls, status_list = [], [], []
        for oci_model_deployment_object in model_deployments:
            state_of_model = State._from_str(
                oci_model_deployment_object.lifecycle_state
            )
            if status == State.UNKNOWN or status.name == state_of_model.name:
                model_dep_id = oci_model_deployment_object.id
                model_dep_url = oci_model_deployment_object._model_deployment_url
                model_dep_state = oci_model_deployment_object.lifecycle_state
                ids.append(model_dep_id)
                urls.append(model_dep_url)
                status_list.append(model_dep_state)
        display["model_id"] = ids
        display["deployment_url"] = urls
        display["current_state"] = status_list
        return display



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from functools import wraps
import json
import time
import logging
from typing import Callable, List
from ads.common.oci_datascience import OCIDataScienceMixin
from ads.common import utils as progress_bar_utils
from ads.config import PROJECT_OCID
from ads.model.deployment.common import utils
from ads.model.deployment.common.utils import OCIClientManager, State
import oci

from oci.data_science.models import (
    CreateModelDeploymentDetails,
    UpdateModelDeploymentDetails,
)

DEFAULT_WAIT_TIME = 1200
DEFAULT_POLL_INTERVAL = 10
DEACTIVATE_WORKFLOW_STEPS = 2
DELETE_WORKFLOW_STEPS = 2
ACTIVATE_WORKFLOW_STEPS = 6
CREATE_WORKFLOW_STEPS = 6
ALLOWED_STATUS = [
    State.ACTIVE.name,
    State.CREATING.name,
    State.DELETED.name,
    State.DELETING.name,
    State.FAILED.name,
    State.INACTIVE.name,
    State.UPDATING.name,
]

logger = logging.getLogger(__name__)

MODEL_DEPLOYMENT_NEEDS_TO_BE_DEPLOYED = "Missing model deployment id. Model deployment needs to be deployed before it can be accessed."


def check_for_model_deployment_id(msg: str = MODEL_DEPLOYMENT_NEEDS_TO_BE_DEPLOYED):
    """The decorator helping to check if the ID attribute sepcified for a datascience model deployment.

    Parameters
    ----------
    msg: str
        The message that will be thrown.

    Raises
    ------
    MissingModelDeploymentIdError
        In case if the ID attribute not specified.

    Examples
    --------
    >>> @check_for_id(msg="Some message.")
    ... def test_function(self, name: str, last_name: str)
    ...     pass
    """

    def decorator(func: Callable):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.id:
                raise MissingModelDeploymentIdError(msg)
            return func(self, *args, **kwargs)

        return wrapper

    return decorator


class MissingModelDeploymentIdError(Exception):
    pass


class MissingModelDeploymentWorkflowIdError(Exception):
    pass


class OCIDataScienceModelDeployment(
    OCIDataScienceMixin,
    oci.data_science.models.ModelDeployment,
):
    """Represents an OCI Data Science Model Deployment.
    This class contains all attributes of the `oci.data_science.models.ModelDeployment`.
    The main purpose of this class is to link the `oci.data_science.models.ModelDeployment`
    and the related client methods.
    Linking the `ModelDeployment` (payload) to Create/Update/Delete/Activate/Deactivate methods.

    The `OCIDataScienceModelDeployment` can be initialized by unpacking the properties stored in a dictionary:

    .. code-block:: python

        properties = {
            "compartment_id": "<compartment_ocid>",
            "name": "<model_name>",
            "description": "<model_description>",
        }
        ds_model_deployment = OCIDataScienceModelDeployment(**properties)

    The properties can also be OCI REST API payload, in which the keys are in camel format.

    .. code-block:: python

        payload = {
            "compartmentId": "<compartment_ocid>",
            "name": "<model_name>",
            "description": "<model_description>",
        }
        ds_model_deployment = OCIDataScienceModelDeployment(**payload)

    Methods
    -------
    activate(self, ...) -> "OCIDataScienceModelDeployment":
        Activates datascience model deployment.
    create(self, ...) -> "OCIDataScienceModelDeployment"
        Creates datascience model deployment.
    deactivate(self, ...) -> "OCIDataScienceModelDeployment":
        Deactivates datascience model deployment.
    delete(self, ...) -> "OCIDataScienceModelDeployment":
        Deletes datascience model deployment.
    update(self, ...) -> "OCIDataScienceModelDeployment":
        Updates datascience model deployment.
    list(self, ...) -> list[oci.data_science.models.ModelDeployment]:
        List oci.data_science.models.ModelDeployment instances within given compartment and project.
    from_id(cls, model_deployment_id: str) -> "OCIDataScienceModelDeployment":
        Gets model deployment by OCID.

    Examples
    --------
    >>> oci_model_deployment = OCIDataScienceModelDeployment.from_id(<model_deployment_ocid>)
    >>> oci_model_deployment.deactivate()
    >>> oci_model_deployment.activate(wait_for_completion=False)
    >>> oci_model_deployment.description = "A brand new description"
    ... oci_model_deployment.update()
    >>> oci_model_deployment.sync()
    >>> oci_model_deployment.list(status="ACTIVE")
    >>> oci_model_deployment.delete(wait_for_completion=False)
    """

    def __init__(
        self,
        config: dict = None,
        signer: oci.signer.Signer = None,
        client_kwargs: dict = None,
        **kwargs,
    ) -> None:
        super().__init__(config, signer, client_kwargs, **kwargs)
        self.workflow_req_id = None

    @check_for_model_deployment_id(
        msg="Model deployment needs to be deployed before it can be activated or deactivated."
    )
    def activate(
        self,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
    ) -> "OCIDataScienceModelDeployment":
        """Activates datascience model deployment.

        Parameters
        ----------
        wait_for_completion: bool
            Flag set for whether to wait for process to be completed.
            Defaults to True.
        max_wait_time: int
            Maximum amount of time to wait in seconds (Defaults to 1200).
            Negative implies infinite wait time.
        poll_interval: int
            Poll interval in seconds (Defaults to 10).

        Returns
        -------
        OCIDataScienceModelDeployment
            The `OCIDataScienceModelDeployment` instance (self).
        """
        logger.info(f"Activating model deployment `{self.id}`.")
        response = self.client.activate_model_deployment(
            self.id,
        )

        if wait_for_completion:

            self.workflow_req_id = response.headers.get("opc-work-request-id", None)
            oci_model_deployment_object = self.client.get_model_deployment(self.id).data
            current_state = State._from_str(oci_model_deployment_object.lifecycle_state)
            model_deployment_id = self.id

            try:
                self._wait_for_progress_completion(
                    State.ACTIVE.name,
                    ACTIVATE_WORKFLOW_STEPS,
                    [State.FAILED.name, State.INACTIVE.name],
                    self.workflow_req_id,
                    current_state,
                    model_deployment_id,
                    max_wait_time,
                    poll_interval,
                )
            except Exception as e:
                logger.error(
                    f"Error while trying to activate model deployment: {self.id}"
                )
                raise e

        return self.sync()

    def create(
        self,
        create_model_deployment_details: CreateModelDeploymentDetails,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
    ) -> "OCIDataScienceModelDeployment":
        """Creates datascience model deployment.

        Parameters
        ----------
        wait_for_completion: bool
            Flag set for whether to wait for process to be completed.
            Defaults to True.
        max_wait_time: int
            Maximum amount of time to wait in seconds (Defaults to 1200).
            Negative implies infinite wait time.
        poll_interval: int
            Poll interval in seconds (Defaults to 10).

        Returns
        -------
        OCIDataScienceModelDeployment
            The `OCIDataScienceModelDeployment` instance (self).
        """
        response = self.client.create_model_deployment(create_model_deployment_details)
        self.update_from_oci_model(response.data)
        logger.info(f"Creating model deployment `{self.id}`.")

        if wait_for_completion:

            self.workflow_req_id = response.headers.get("opc-work-request-id", None)
            res_payload = json.loads(str(response.data))
            current_state = State._from_str(res_payload["lifecycle_state"])
            model_deployment_id = self.id

            try:
                self._wait_for_progress_completion(
                    State.ACTIVE.name,
                    CREATE_WORKFLOW_STEPS,
                    [State.FAILED.name, State.INACTIVE.name],
                    self.workflow_req_id,
                    current_state,
                    model_deployment_id,
                    max_wait_time,
                    poll_interval,
                )
            except Exception as e:
                logger.error(
                    f"Error while trying to create model deployment: {self.id}"
                )
                raise e

        return self.sync()

    @check_for_model_deployment_id(
        msg="Model deployment needs to be deployed before it can be activated or deactivated."
    )
    def deactivate(
        self,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
    ) -> "OCIDataScienceModelDeployment":
        """Deactivates datascience model deployment.

        Parameters
        ----------
        wait_for_completion: bool
            Flag set for whether to wait for process to be completed.
            Defaults to True.
        max_wait_time: int
            Maximum amount of time to wait in seconds (Defaults to 1200).
            Negative implies infinite wait time.
        poll_interval: int
            Poll interval in seconds (Defaults to 10).

        Returns
        -------
        OCIDataScienceModelDeployment
            The `OCIDataScienceModelDeployment` instance (self).
        """
        logger.info(f"Deactivating model deployment `{self.id}`.")
        response = self.client.deactivate_model_deployment(
            self.id,
        )

        if wait_for_completion:

            self.workflow_req_id = response.headers.get("opc-work-request-id", None)
            oci_model_deployment_object = self.client.get_model_deployment(self.id).data
            current_state = State._from_str(oci_model_deployment_object.lifecycle_state)
            model_deployment_id = self.id

            try:
                self._wait_for_progress_completion(
                    State.INACTIVE.name,
                    DEACTIVATE_WORKFLOW_STEPS,
                    [State.FAILED.name],
                    self.workflow_req_id,
                    current_state,
                    model_deployment_id,
                    max_wait_time,
                    poll_interval,
                )
            except Exception as e:
                logger.error(
                    f"Error while trying to deactivate model deployment: {self.id}"
                )
                raise e

        return self.sync()

    @check_for_model_deployment_id(
        msg="Model deployment needs to be deployed before it can be deleted."
    )
    def delete(
        self,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
    ) -> "OCIDataScienceModelDeployment":
        """Deletes datascience model deployment.

        Parameters
        ----------
        wait_for_completion: bool
            Flag set for whether to wait for process to be completed.
            Defaults to True.
        max_wait_time: int
            Maximum amount of time to wait in seconds (Defaults to 1200).
            Negative implies infinite wait time.
        poll_interval: int
            Poll interval in seconds (Defaults to 10).

        Returns
        -------
        OCIDataScienceModelDeployment
            The `OCIDataScienceModelDeployment` instance (self).
        """
        logger.info(f"Deleting model deployment `{self.id}`.")
        response = self.client.delete_model_deployment(
            self.id,
        )

        if wait_for_completion:

            self.workflow_req_id = response.headers.get("opc-work-request-id", None)
            oci_model_deployment_object = self.client.get_model_deployment(self.id).data
            current_state = State._from_str(oci_model_deployment_object.lifecycle_state)
            model_deployment_id = self.id

            try:
                self._wait_for_progress_completion(
                    State.DELETED.name,
                    DELETE_WORKFLOW_STEPS,
                    [State.FAILED.name, State.INACTIVE.name],
                    self.workflow_req_id,
                    current_state,
                    model_deployment_id,
                    max_wait_time,
                    poll_interval,
                )
            except Exception as e:
                logger.error(
                    f"Error while trying to delete model deployment: {self.id}"
                )
                raise e

        return self.sync()

    @check_for_model_deployment_id(
        msg="Model deployment needs to be deployed before it can be updated."
    )
    def update(
        self,
        update_model_deployment_details: UpdateModelDeploymentDetails,
        wait_for_completion: bool = True,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
    ) -> "OCIDataScienceModelDeployment":
        """Updates datascience model deployment.

        Parameters
        ----------
        update_model_deployment_details: UpdateModelDeploymentDetails
            Details to update model deployment.
        wait_for_completion: bool
            Flag set for whether to wait for process to be completed.
            Defaults to True.
        max_wait_time: int
            Maximum amount of time to wait in seconds (Defaults to 1200).
            Negative implies infinite wait time.
        poll_interval: int
            Poll interval in seconds (Defaults to 10).

        Returns
        -------
        OCIDataScienceModelDeployment
            The `OCIDataScienceModelDeployment` instance (self).
        """
        if wait_for_completion:
            wait_for_states = [
                oci.data_science.models.WorkRequest.STATUS_SUCCEEDED,
                oci.data_science.models.WorkRequest.STATUS_FAILED,
            ]
        else:
            wait_for_states = []

        try:
            response = self.client_composite.update_model_deployment_and_wait_for_state(
                self.id,
                update_model_deployment_details,
                wait_for_states=wait_for_states,
                waiter_kwargs={
                    "max_interval_seconds": poll_interval,
                    "max_wait_seconds": max_wait_time,
                },
            )
            self.workflow_req_id = response.headers.get("opc-work-request-id", None)
        except Exception as e:
            logger.error(f"Error while trying to update model deployment: {self.id}")
            raise e

        return self.sync()

    @classmethod
    def list(
        cls,
        status: str = None,
        compartment_id: str = None,
        project_id: str = None,
        **kwargs,
    ) -> list:
        """Lists the model deployments associated with current compartment id and status

        Parameters
        ----------
        status : str
            Status of deployment. Defaults to None.
            Allowed values: `ACTIVE`, `CREATING`, `DELETED`, `DELETING`, `FAILED`, `INACTIVE` and `UPDATING`.
        compartment_id : str
            Target compartment to list deployments from.
            Defaults to the compartment set in the environment variable "NB_SESSION_COMPARTMENT_OCID".
            If "NB_SESSION_COMPARTMENT_OCID" is not set, the root compartment ID will be used.
            An ValueError will be raised if root compartment ID cannot be determined.
        project_id : str
            Target project to list deployments from.
            Defaults to the project id in the environment variable "PROJECT_OCID".
        kwargs :
            The values are passed to oci.data_science.DataScienceClient.list_model_deployments.

        Returns
        -------
        list
            A list of oci.data_science.models.ModelDeployment objects.

        Raises
        ------
        ValueError
            If compartment_id is not specified and cannot be determined from the environment.
        """
        compartment_id = compartment_id or OCIClientManager().default_compartment_id()

        if not compartment_id:
            raise ValueError(
                "Unable to determine compartment ID from environment. Specify `compartment_id`."
            )

        project_id = project_id or PROJECT_OCID
        if project_id:
            kwargs["project_id"] = project_id

        if status is not None:
            if status not in ALLOWED_STATUS:
                raise ValueError(
                    f"Allowed `status` values are: {', '.join(ALLOWED_STATUS)}."
                )
            kwargs["lifecycle_state"] = status

        # https://oracle-cloud-infrastructure-python-sdk.readthedocs.io/en/latest/api/pagination.html#module-oci.pagination
        return oci.pagination.list_call_get_all_results(
            cls().client.list_model_deployments, compartment_id, **kwargs
        ).data

    @classmethod
    def from_id(cls, model_deployment_id: str) -> "OCIDataScienceModelDeployment":
        """Gets datascience model deployment by OCID.

        Parameters
        ----------
        model_deployment_id: str
            The OCID of the datascience model deployment.

        Returns
        -------
        OCIDataScienceModelDeployment
            An instance of `OCIDataScienceModelDeployment`.
        """
        return super().from_ocid(model_deployment_id)

    def _wait_for_progress_completion(
        self,
        final_state: str,
        work_flow_step: int,
        disallowed_final_states: List[str],
        work_flow_request_id: str,
        state: State,
        model_deployment_id: str,
        max_wait_time: int = DEFAULT_WAIT_TIME,
        poll_interval: int = DEFAULT_POLL_INTERVAL,
    ):
        """_wait_for_progress_completion blocks until progress is completed.

        Parameters
        ----------
        final_state: str
            Final state of model deployment aimed to be reached.
        work_flow_step: int
            Number of work flow step of the request.
        disallowed_final_states: list[str]
            List of disallowed final state to be reached.
        work_flow_request_id: str
            The id of work flow request.
        state: State
            The current state of model deployment.
        model_deployment_id: str
            The ocid of model deployment.
        max_wait_time: int
            Maximum amount of time to wait in seconds (Defaults to 1200).
            Negative implies infinite wait time.
        poll_interval: int
            Poll interval in seconds (Defaults to 10).
        """

        start_time = time.time()
        prev_message = ""
        prev_workflow_stage_len = 0
        current_state = state or State.UNKNOWN
        with progress_bar_utils.get_progress_bar(work_flow_step) as progress:
            if max_wait_time > 0 and utils.seconds_since(start_time) >= max_wait_time:
                utils.get_logger().error(
                    f"Max wait time ({max_wait_time} seconds) exceeded."
                )
            while (
                max_wait_time < 0 or utils.seconds_since(start_time) < max_wait_time
            ) and current_state.name.upper() != final_state:
                if current_state.name.upper() in disallowed_final_states:
                    utils.get_logger().info(
                        f"Operation failed due to deployment reaching state {current_state.name.upper()}. Use Deployment ID for further steps."
                    )
                    break

                prev_state = current_state.name
                try:
                    model_deployment_payload = json.loads(
                        str(self.client.get_model_deployment(model_deployment_id).data)
                    )
                    current_state = (
                        State._from_str(model_deployment_payload["lifecycle_state"])
                        if "lifecycle_state" in model_deployment_payload
                        else State.UNKNOWN
                    )
                    workflow_payload = self.client.list_work_request_logs(
                        work_flow_request_id
                    ).data
                    if isinstance(workflow_payload, list) and len(workflow_payload) > 0:
                        if prev_message != workflow_payload[-1].message:
                            for _ in range(
                                len(workflow_payload) - prev_workflow_stage_len
                            ):
                                progress.update(workflow_payload[-1].message)
                            prev_workflow_stage_len = len(workflow_payload)
                            prev_message = workflow_payload[-1].message
                            prev_workflow_stage_len = len(workflow_payload)
                    if prev_state != current_state.name:
                        utils.get_logger().info(
                            f"Status Update: {current_state.name} in {utils.seconds_since(start_time)} seconds"
                        )
                except Exception as e:
                    # utils.get_logger().warning(
                    #     "Unable to update deployment status. Details: %s", format(
                    #         e)
                    # )
                    pass
                time.sleep(poll_interval)
            progress.update("Done")



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
"""Utilities used by the model deployment package
"""

# Standard lib

import json
import logging
import os
import shutil
import tempfile
import time
import requests
from typing import Dict
from enum import Enum, auto

import oci
import fsspec
from oci.data_science.models import CreateModelDetails
from ads.common import auth, oci_client
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.config import COMPARTMENT_OCID, PROJECT_OCID


logger = logging.getLogger(__name__)
DEFAULT_CONTENT_TYPE_JSON = "application/json"
DEFAULT_CONTENT_TYPE_BYTES = "application/octet-stream"


def get_logger():
    return logger


def set_log_level(level="INFO"):
    """set_log_level sets the logger level

    Args:
        level (str, optional): The logger level. Defaults to "INFO"

    Returns:
        Nothing
    """

    level = logging.getLevelName(level)
    logger.setLevel(level)


def seconds_since(t):
    """seconds_since returns the seconds since `t`. `t` is assumed to be a time
    in epoch seconds since time.time() returns the current time in epoch seconds.

    Args:
        t (int) - a time in epoch seconds

    Returns
        int: the number of seconds since `t`
    """

    return time.time() - t


@runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
def is_notebook():
    """is_notebook returns True if the environment is a Jupyter notebook and
    False otherwise

    Args:
        None

    Returns:
        bool: True if Jupyter notebook; False otherwise

    Raises:
        NameError: If retrieving the shell name from get_ipython() throws an error

    """

    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":  # pragma: no cover
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except ImportError:
        # IPython is not installed
        return False
    except NameError:
        return False  # Probably standard Python interpreter


def send_request(
    data,
    endpoint: str,
    dry_run: bool = False,
    is_json_payload: bool = False,
    header: dict = {},
    **kwargs,
):
    """Sends the data to the predict endpoint.

    Args:
        data (bytes or Json serializable): data need to be sent to the endpoint.
        endpoint (str): The model HTTP endpoint.
        dry_run (bool, optional): Defaults to False.
        is_json_payload (bool, optional): Indicate whether to send data with a `application/json` MIME TYPE. Defaults to False.
        header (dict, optional): A dictionary of HTTP headers to send to the specified url. Defaults to {}.

    Returns:
        A JSON representive of a requests.Response object.
    """
    headers = dict()
    if is_json_payload:
        headers["Content-Type"] = (
            header.get("content_type") or DEFAULT_CONTENT_TYPE_JSON
        )
        request_kwargs = {"json": data}
    else:
        headers["Content-Type"] = (
            header.get("content_type") or DEFAULT_CONTENT_TYPE_BYTES
        )
        request_kwargs = {"data": data}  # should pass bytes when using data

    request_kwargs["headers"] = headers

    if dry_run:
        request_kwargs["headers"]["Accept"] = "*/*"
        req = requests.Request("POST", endpoint, **request_kwargs).prepare()
        if is_json_payload:
            return json.loads(req.body)
        return req.body
    else:
        request_kwargs["auth"] = header.get("signer")
        return requests.post(endpoint, **request_kwargs).json()


# State Constants
class State(Enum):
    ACTIVE = auto()
    CREATING = auto()
    DELETED = auto()
    DELETING = auto()
    FAILED = auto()
    INACTIVE = auto()
    UPDATING = auto()
    UNKNOWN = auto()

    @staticmethod
    def _from_str(state):
        if state == None:
            return State.UNKNOWN
        elif state.upper() == "ACTIVE":
            return State.ACTIVE
        elif state.upper() == "CREATING":
            return State.CREATING
        elif state.upper() == "DELETED":
            return State.DELETED
        elif state.upper() == "DELETING":
            return State.DELETING
        elif state.upper() == "FAILED":
            return State.FAILED
        elif state.upper() == "INACTIVE":
            return State.INACTIVE
        elif state.upper() == "UPDATING":
            return State.UPDATING
        else:
            return State.UNKNOWN

    def __call__(self):
        # This will provide backward compatibility.
        # In previous release, ModelDeployment has state() as method instead of property
        return self


class OCIClientManager:
    """OCIClientManager is a helper class used for accessing DataScienceClient and
    DataScienceCompositeClient objects

    Attributes
    ----------
    ds_client - class attribute for data science client
    ds_composite_client - class attribute for data science composite client

    """

    def __init__(self, config=None) -> None:
        if not config:
            config = auth.default_signer()
        self.config = config
        self.ds_client = oci_client.OCIClientFactory(**config).data_science
        self.ds_composite_client = (
            oci.data_science.DataScienceClientCompositeOperations(self.ds_client)
        )

    def default_compartment_id(self):
        """Determines the default compartment OCID
        This method finds the compartment OCID from (in priority order):
        an environment variable, an API key config or a resource principal signer.

        Parameters
        ----------
        config : dict, optional
            The model deployment config, which contains the following keys:
            auth: Authentication method, must be either "resource_principal" or "api_key".
            If auth is not specified:
                1. api_key will be used if available.
                2. If api_key is not available, resource_principal will be used.
            oci_config_file: OCI API key config file location. Defaults to "~/.oci/config"
            oci_config_profile: OCI API key config profile name. Defaults to "DEFAULT"

        Returns
        -------
        str or None
            The compartment OCID if found. Otherwise None.
        """
        # Try to get compartment ID from environment variable.')
        if os.environ.get("NB_SESSION_COMPARTMENT_OCID"):
            return os.environ.get("NB_SESSION_COMPARTMENT_OCID")
        # Try to get compartment ID from OCI config, then RP signer
        # Note: we assume compartment_ids can never be: 0, False, etc.
        oci_config = self.config.get("config")
        signer = self.config.get("signer")
        return (
            oci_config.get("compartment_id")
            or oci_config.get("tenancy")
            or getattr(signer, "tenancy_id", None)
        )

    def prepare_artifact(self, model_uri: str, properties: Dict) -> str:
        """
        Prepare model artifact. Returns model ocid.

        Args:
            model_uri (str): uri to model files, can be local or in cloud storage
            properties (dict): dictionary of properties that are needed for creating a model.
            ds_client (DataScienceClient): OCI DataScienceClient

        Returns:
            str: model ocid
        """
        properties_dict = {}
        if properties:
            properties_dict = (
                properties
                if isinstance(properties, dict)
                else json.loads(repr(properties))
            )
        with tempfile.TemporaryDirectory() as d:
            fhandlers = fsspec.open_files(
                model_uri,
                config=self.config.get("config", {}),
                mode="rb",
            )
            if len(fhandlers) == 0:
                raise FileNotFoundError("No files found under this path.")
            for fh in fhandlers:
                with fh as fin:
                    with open(os.path.join(d, os.path.basename(fh.path)), "wb") as fout:
                        fout.write(fin.read())
            shutil.make_archive(
                os.path.join(os.path.dirname(d), "model_files"),
                "zip",
                os.path.dirname(d),
                os.path.basename(d),
            )
            return self._upload_artifact(
                f"{os.path.join(os.path.dirname(d), 'model_files')}.zip",
                properties_dict,
            )

    def _upload_artifact(self, model_zip: str, properties: dict) -> str:
        """Uploads the model artifact to cloud storage.

        Args:
            ds_client (DataScienceClient): OCI DataScienceClient
            model_zip (str): path to model artifact zip file
            properties (dict): dictionary of properties

        Returns:
            str: model ocid
        """
        project_id = properties.get("project_id", PROJECT_OCID)
        compartment_id = properties.get("compartment_id", COMPARTMENT_OCID)

        if not project_id or not compartment_id:
            raise ValueError(
                "Both `project_id` and `compartment_id` need to be provided. You can pass them through kwargs or `ModelDeploymentProperties` object."
            )

        create_model_details = CreateModelDetails(
            display_name=properties.get("display_name", None),
            project_id=project_id,
            compartment_id=compartment_id,
        )

        model = self.ds_client.create_model(create_model_details).data
        with open(model_zip, "rb") as data:
            self.ds_client.create_model_artifact(
                model.id,
                data,
                content_disposition=f'attachment; filename="{model.id}.zip"',
            )
        return model.id
